
"""
****** Thevenaz Spline Convolution Implementation ****************************************
Thevenaz, Philippe, Thierry Blu, and Michael Unser. ?Image Interpolation and Resampling?, n.d., 39.
http://bigwww.epfl.ch/thevenaz/interpolation/
*******************************************************************************
"""

from maptial.xyz import vectorthree as v3
import math
import numpy as np

#from . import invariant as ivm
from . import iv1
from . import iv2
from . import iv3
from . import iv5

import numpy as np

class Thevenaz(object):
    def __init__(self,values, degree,log_level=0):
        self.shape = values.shape
        self.dims = len(self.shape)
        self.degree = degree
        self.log_level = log_level         
        self.values = values.astype(float)
        if log_level > 0:
            print("Thevenaz Interpolator:",self.degree,self.shape,self.dims)    
        self._buffer = 2        
        self.h = 0.0001 #this is the interval for numerical differentiation       
        self.TOLERANCE = 2.2204460492503131e-016 # smallest such that 1.0+DBL_EPSILON != 1.0                                
        self.make_mirror_coeffs()         
    ############################################################################################################    
    def get_radient_numerical(self, x, y, z):                        
        val = self.get_value(x, y, z)
        dx = (self.get_value(x + self.h, y, z) - val) / self.h
        dy = (self.get_value(x, y + self.h, z) - val) / self.h
        dz = (self.get_value(x, y, z + self.h) - val) / self.h
        radient = (abs(dx) + abs(dy) + abs(dz)) / 3
        return radient
    ############################################################################################################    
    def get_laplacian_numerical(self, x, y, z):        
        val = self.get_value(x, y, z)
        xx = self.getDxDx_numerical(x, y, z, val)
        yy = self.getDyDy_numerical(x, y, z, val)
        zz = self.getDzDz_numerical(x, y, z, val)
        return xx + yy + zz 
    ############################################################################################################    
    def getDxDx_numerical(self, x, y, z, val):        
        va = self.get_value(x - self.h, y, z)
        vb = self.get_value(x + self.h, y, z)
        dd = (va + vb - 2 * val) / (self.h * self.h)
        return dd
    ############################################################################################################    
    def getDyDy_numerical(self, x, y, z, val):        
        va = self.get_value(x, y - self.h, z)
        vb = self.get_value(x, y + self.h, z)
        dd = (va + vb - 2 * val) / (self.h * self.h)
        return dd
    ############################################################################################################    
    def getDzDz_numerical(self, x, y, z, val):        
        va = self.get_value(x, y, z - self.h)
        vb = self.get_value(x, y, z + self.h)
        dd = (va + vb - 2 * val) / (self.h * self.h)
        return dd
    ############################################################################################################    
    def get_periodic_value(self,coords):
        new_coords = np.copy(coords)
        for c in len(coords):
            coord = coords[c]
            limit = self.shape[c]
            while coord < 0:
                coord = limit + coord
            while coord >= limit:
                coord = coord - limit
            new_coords[c] = coord        
        return self._npy[new_coords]
    ############################################################################################################    
    def get_exact_value(self,coords):        
        return self._npy[coords]
    ############################################################################################################                                    
    def get_radient(self, x, y, z):
        return self.get_radient_numerical(x,y,z)
    ############################################################################################################                                    
    def get_laplacian(self, x, y, z):
        return self.get_laplacian_numerical(x,y,z)
    ############################################################################################################                                            
    def make_mirror_coeffs(self):
        self.coeffs = np.copy(self.values)
        if len(self.shape) == 3:
            self.make_mirror_coeffs_3d(self.coeffs)
        elif len(self.shape) == 2:
            self.make_mirror_coeffs_2d(self.coeffs)
        elif len(self.shape) == 1:
            self.make_mirror_coeffs_1d(self.coeffs)

    ############################################################################################################                                            
    def make_mirror_coeffs_3d(self,values):
        shp = values.shape
        new_shp = (shp[0],shp[1])
        for i in range(values.shape[2]):
            vals = values[:,:,i:i+1].reshape(new_shp)
            self.make_mirror_coeffs_2d(vals)
    ############################################################################################################                                            
    def make_mirror_coeffs_2d(self,values):
        shp = values.shape
        new_shp = (shp[0])
        for i in range(values.shape[1]):
            vals = values[:,i:i+1].reshape(new_shp)
            self.make_mirror_coeffs_1d(vals)
    ############################################################################################################                                            
    def make_mirror_coeffs_1d(self,values):        
        pole = self.get_pole(self.degree)
        num_poles = len(pole)
        self.convert_to_interp_coeffs(pole, num_poles, len(values), values)
    ############################################################################################################                                                                                                                                                        
    def get_pole(self,degree):        
        #Recover the poles from a lookup table #currently only 3 degree, will I want to calculate all the possibilities at the beginnning, 3,5,7,9?
        pole = []
        if (degree == 9):        
            pole.append(-0.60799738916862577900772082395428976943963471853991)
            pole.append(-0.20175052019315323879606468505597043468089886575747)
            pole.append(-0.043222608540481752133321142979429688265852380231497)
            pole.append(-0.0021213069031808184203048965578486234220548560988624)
        elif (degree == 7):        
            pole.append(-0.53528043079643816554240378168164607183392315234269)
            pole.append(-0.12255461519232669051527226435935734360548654942730)
            pole.append(-0.0091486948096082769285930216516478534156925639545994)
        elif (degree == 5):        
            pole.append(np.sqrt(135.0 / 2.0 - np.sqrt(17745.0 / 4.0)) + np.sqrt(105.0 / 4.0) - 13.0 / 2.0)
            pole.append(np.sqrt(135.0 / 2.0 + np.sqrt(17745.0 / 4.0)) - np.sqrt(105.0 / 4.0) - 13.0 / 2.0)
        else:#then it is 3        
            pole.append(np.sqrt(3.0) - 2.0)
        return pole
    ############################################################################################################                                                
    def convert_to_interp_coeffs(self,pole, num_poles, width, row,mirror=True):        
        #/* special case required by mirror boundaries */
        if (width == 1):         
            #mirror filter and periodic filter, not much can be done if it is only 1 think
            return row        
        lmbda = 1
        n = 0
        k = 0
        #Compute the overall gain
        for k in range(num_poles):
            lmbda = lmbda * (1 - pole[k]) * (1 - 1 / pole[k])        
        #Apply the gain
        for n in range(width):
            row[n] = float(row[n]) * float(lmbda)        
        #loop over the poles            
        for k in range(num_poles):        
            #/* causal initialization */                        
            row[0] = self.initial_causal_coeffs(row, width, pole[k])            
            #/* causal recursion */
            for n in range(1,width):            
                row[n] = row[n]+ pole[k] * row[n - 1]
            #/* anticausal initialization */            
            row[width - 1] = self.initial_anticausal_coeffs(row, width, pole[k])            
            #/* anticausal recursion */
            for n in range(width - 2, -1, -1):
                row[n] = pole[k] * (row[n + 1] - row[n])
        return row
    ############################################################################################################                                                
    def initial_causal_coeffs(self, vals, length, pole):
        #/* begin InitialCausalCoefficient */
        Sum, zn, z2n, iz = 0.,0.,0.,0.
        n, Horizon = 0,0
        #/* this initialization corresponds to mirror boundaries */
        Horizon = length
        if (self.TOLERANCE > 0.0):        
            Horizon = np.ceil(np.log(self.TOLERANCE) / np.log(abs(pole)))
        if (Horizon < length):        
            #/* accelerated loop */
            zn = pole
            Sum = vals[0]
            for n in range(1,int(Horizon)):
                Sum += zn * vals[n]
                zn *= pole       
            return Sum        
        else:#// if (_mirror)
            #// RSA notes this is the mirror condition, when the horizon - how far you need to look ahead for good data - is not as far the data you have
            #/* full loop */
            zn = pole
            iz = 1.0 / pole
            z2n = math.pow(pole, length - 1)
            Sum = vals[0] + z2n * vals[length - 1]
            z2n *= z2n * iz 
            for n in range(1,length - 1):
                Sum += (zn + z2n) * vals[n]
                zn *= pole                
                z2n *= iz
            return (Sum / (1.0 - zn * zn))        
    ############################################################################################################                                                        
    def initial_anticausal_coeffs(self, vals, length, pole):        
        #/* this initialization corresponds to mirror boundaries */
        if (length < 2):
            return 0;
        else:#// if (_mirror)
            return ((pole / (pole * pole - 1.0)) * (pole * vals[length - 2] + vals[length - 1]))
    ############################################################################################################                                            
    def get_coeff(self,xyz):
        try:                        
            value = self.coeffs[xyz]
        except Exception as e:
            print(xyz,str(e))

        return value
    ############################################################################################################                                                                                
    def put_coeff(self,xyz,v):
        self.coeffs[xyz] = v
    ############################################################################################################                                            
    def get_value(self,x,y=None,z=None):
        if len(self.shape) == 1:
            return self.get_value1d(x)
        elif len(self.shape) == 2:
            return self.get_value2d(x,y)
        elif len(self.shape) == 3:
            return self.get_value3d(x,y,z)
    ############################################################################################################                                            
    def get_value1d(self, x):
        weight_length = self.degree + 1
        xIndex = []        
        xWeight = []        
        for s in range(weight_length):            
            xIndex.append(0)                        
            xWeight.append(0)
        i = int(np.floor(x) - np.floor(self.degree / 2))
        for l in range(self.degree+1):                          
            xIndex[l] = i #if 71.1 passed in, for linear, we would want 71 and 72,             
            i=i+1
        if (self.degree == 9):        
            xWeight = self.applyValue9(x, xIndex, weight_length)            
        elif (self.degree == 7):        
            xWeight = self.applyValue7(x, xIndex, weight_length)            
        elif (self.degree == 5):        
            xWeight = self.applyValue5(x, xIndex, weight_length)            
        else:        
            xWeight = self.applyValue3(x, xIndex, weight_length)        
        
        # !!! I have removed the mirror boundary bit, might need to put it back in RSA TODO                                  
        Width2 = 2 * self.shape[0] - 2        
        mirror_convert = [(xIndex,Width2,self.shape[0])]
        for pIndex,pLen2,pLen in mirror_convert:            
            for k in range(len(pIndex)):        
                if pLen == 1:
                    pIndex[k] = 0
                elif pIndex[k] < 0:
                    pIndex[k] = int(-pIndex[k] - pLen2 * ((-pIndex[k]) / pLen2))                
                elif pLen <= pIndex[k]:
                    pIndex[k] = int(pLen2 - pIndex[k])
        w1 = 0.0        
        for i in range(self.degree+1):                
            w1 += xWeight[i] * self.get_coeff(xIndex[i])        
        return w1
    
    def get_value2d(self, x,y):
        weight_length = self.degree + 1
        xIndex,yIndex = [],[]
        xWeight,yWeight = [],[]
        for s in range(weight_length):            
            xIndex.append(0)                        
            xWeight.append(0)
            yIndex.append(0)                        
            yWeight.append(0)
        i = int(np.floor(x) - np.floor(self.degree / 2))
        j = int(np.floor(y) - np.floor(self.degree / 2))
        for l in range(self.degree+1):                          
            xIndex[l] = i #if 71.1 passed in, for linear, we would want 71 and 72,             
            yIndex[l] = j 
            i,j=i+1,j+1
        if (self.degree == 9):        
            xWeight = self.applyValue9(x, xIndex, weight_length)            
            yWeight = self.applyValue9(y, yIndex, weight_length)            
        elif (self.degree == 7):        
            xWeight = self.applyValue7(x, xIndex, weight_length)            
            yWeight = self.applyValue7(y, yIndex, weight_length)            
        elif (self.degree == 5):        
            xWeight = self.applyValue5(x, xIndex, weight_length)            
            yWeight = self.applyValue5(y, yIndex, weight_length)            
        else:        
            xWeight = self.applyValue3(x, xIndex, weight_length)        
            yWeight = self.applyValue3(y, yIndex, weight_length)            
        
        # !!! I have removed the mirror boundary bit, might need to put it back in RSA TODO                          
        Width2 = 2 * self.shape[0] - 2
        Height2 = 2 * self.shape[1] - 2        
        mirror_convert = [(xIndex,Width2,self.shape[0]),(yIndex,Height2,self.shape[1])]
        for pIndex,pLen2,pLen in mirror_convert:
            for k in range(len(pIndex)):        
                if pLen == 1:
                    pIndex[k] = 0
                elif pIndex[k] < 0:
                    pIndex[k] = int(-pIndex[k] - pLen2 * ((-pIndex[k]) / pLen2))                
                elif pLen <= pIndex[k]:
                    pIndex[k] = int(pLen2 - pIndex[k])
        
        w2,w1 = 0.0,0.0
        for j in range(self.degree+1):
            for i in range(self.degree+1):                
                w1 += xWeight[i] * self.get_coeff((xIndex[i],yIndex[j]))
            w2 += yWeight[j] * w1
        return w2
            

    def get_value3d(self, x,y,z):                
        weight_length = self.degree + 1
        xIndex, yIndex, zIndex = [],[],[]        
        xWeight, yWeight,zWeight = [],[],[]        
        for s in range(weight_length):            
            xIndex.append(0)
            yIndex.append(0)
            zIndex.append(0)
            xWeight.append(0)
            yWeight.append(0)
            zWeight.append(0)            
        #Compute the interpolation indices
        i = int(np.floor(x) - np.floor(self.degree / 2))
        j = int(np.floor(y) - np.floor(self.degree / 2))
        k = int(np.floor(z) - np.floor(self.degree / 2))

        for l in range(self.degree+1):                          
            xIndex[l] = i #if 71.1 passed in, for linear, we would want 71 and 72, 
            yIndex[l] = j
            zIndex[l] = k
            i,j,k=i+1,j+1,k+1              
        #/* compute the interpolation weights */
        if (self.degree == 9):        
            xWeight = self.applyValue9(x, xIndex, weight_length)
            yWeight = self.applyValue9(y, yIndex, weight_length)
            zWeight = self.applyValue9(z, zIndex, weight_length)        
        elif (self.degree == 7):        
            xWeight = self.applyValue7(x, xIndex, weight_length)
            yWeight = self.applyValue7(y, yIndex, weight_length)
            zWeight = self.applyValue7(z, zIndex, weight_length)        
        elif (self.degree == 5):        
            xWeight = self.applyValue5(x, xIndex, weight_length)
            yWeight = self.applyValue5(y, yIndex, weight_length)
            zWeight = self.applyValue5(z, zIndex, weight_length)        
        else:        
            xWeight = self.applyValue3(x, xIndex, weight_length)
            yWeight = self.applyValue3(y, yIndex, weight_length)
            zWeight = self.applyValue3(z, zIndex, weight_length)        
        #//applying the mirror boundary condition becaue I am only interpolating within values??
        #// RSA edit actually I want to wrap        
        # !!! I have removed the mirror boundary bit, might need to put it back in RSA TODO                          
        Width2 = 2 * self.shape[0] - 2
        Height2 = 2 * self.shape[1] - 2
        Depth2 = 2 * self.shape[2] - 2                        
        mirror_convert = [(xIndex,Width2,self.shape[0]),(yIndex,Height2,self.shape[1]),(zIndex,Depth2,self.shape[2])]        
        for pIndex,pLen2,pLen in mirror_convert:
            for k in range(len(pIndex)):        
                if pLen == 1:
                    pIndex[k] = 0
                elif pIndex[k] < 0:
                    pIndex[k] = int(-pIndex[k] - pLen2 * ((-pIndex[k]) / pLen2))                
                elif pLen <= pIndex[k]:
                    pIndex[k] = int(pLen2 - pIndex[k])
        #Perform interolation            
        spline_degree = self.degree
        w3 = 0.0
        for k in range(spline_degree+1):
            w2 = 0.0
            for j in range(spline_degree+1):
                w1 = 0.0
                for i in range(spline_degree+1):                
                    w1 += xWeight[i] * self.get_coeff((xIndex[i],yIndex[j],zIndex[k]))
                w2 += yWeight[j] * w1
            w3 += zWeight[k] * w2
        return w3
    ############################################################################################################                                            
    def applyValue3(self,val, idc, weight_length):        
        ws = []
        for i in range(weight_length):
            ws.append(0)
        w = val - idc[1]        
        ws[3] = (1.0 / 6.0) * w * w * w
        ws[0] = (1.0 / 6.0) + (1.0 / 2.0) * w * (w - 1.0) - ws[3]        
        ws[2] = w + ws[0] - 2.0 * ws[3]
        ws[1] = 1.0 - ws[0] - ws[2] - ws[3]
        
        return ws
    ############################################################################################################                                                
    def applyValue5(self,val,idc,weight_length):        
        ws = []
        for i in range(weight_length):
            ws.append(0)
        w = val - idc[2]
        w2 = w * w
        ws[5] = (1.0 / 120.0) * w * w2 * w2
        w2 -= w
        w4 = w2 * w2
        w -= 1.0 / 2.0
        t = w2 * (w2 - 3.0)
        ws[0] = (1.0 / 24.0) * (1.0 / 5.0 + w2 + w4) - ws[5]
        t0 = (1.0 / 24.0) * (w2 * (w2 - 5.0) + 46.0 / 5.0)
        t1 = (-1.0 / 12.0) * w * (t + 4.0)
        ws[2] = t0 + t1
        ws[3] = t0 - t1
        t0 = (1.0 / 16.0) * (9.0 / 5.0 - t)
        t1 = (1.0 / 24.0) * w * (w4 - w2 - 5.0)
        ws[1] = t0 + t1
        ws[4] = t0 - t1
        return ws
    ############################################################################################################                                                
    def applyValue7(self,val,idc, weight_length):        
        ws = []
        for i in range(weight_length):
            ws.append(0)
        w = val - idc[3]
        ws[0] = 1.0 - w
        ws[0] *= ws[0]
        ws[0] *= ws[0] * ws[0]
        ws[0] *= (1.0 - w) / 5040.0
        w2 = w * w
        ws[1] = (120.0 / 7.0 + w * (-56.0 + w * (72.0 + w * (-40.0 + w2 * (12.0 + w * (-6.0 + w)))))) / 720.0
        ws[2] = (397.0 / 7.0 - w * (245.0 / 3.0 + w * (-15.0 + w * (-95.0 / 3.0 + w * (15.0 + w * (5.0 + w * (-5.0 + w))))))) / 240.0
        ws[3] = (2416.0 / 35.0 + w2 * (-48.0 + w2 * (16.0 + w2 * (-4.0 + w)))) / 144.0
        ws[4] = (1191.0 / 35.0 - w * (-49.0 + w * (-9.0 + w * (19.0 + w * (-3.0 + w) * (-3.0 + w2))))) / 144.0
        ws[5] = (40.0 / 7.0 + w * (56.0 / 3.0 + w * (24.0 + w * (40.0 / 3.0 + w2 * (-4.0 + w * (-2.0 + w)))))) / 240.0
        ws[7] = w2
        ws[7] *= ws[7] * ws[7]
        ws[7] *= w / 5040.0
        ws[6] = 1.0 - ws[0] - ws[1] - ws[2] - ws[3] - ws[4] - ws[5] - ws[7]
        return ws
    ############################################################################################################                                                
    def applyValue9(self,val,idc,weight_length):
        ws = []
        for i in range(weight_length):
            ws.append(0)
        w = val - idc[4]
        ws[0] = 1.0 - w
        ws[0] *= ws[0]
        ws[0] *= ws[0]
        ws[0] *= ws[0] * (1.0 - w) / 362880.0
        ws[1] = (502.0 / 9.0 + w * (-246.0 + w * (472.0 + w * (-504.0 + w * (308.0 + w * (-84.0 + w * (-56.0 / 3.0 + w * (24.0 + w * (-8.0 + w))))))))) / 40320.0
        ws[2] = (3652.0 / 9.0 - w * (2023.0 / 2.0 + w * (-952.0 + w * (938.0 / 3.0 + w * (112.0 + w * (-119.0 + w * (56.0 / 3.0 + w * (14.0 + w * (-7.0 + w))))))))) / 10080.0
        ws[3] = (44117.0 / 42.0 + w * (-2427.0 / 2.0 + w * (66.0 + w * (434.0 + w * (-129.0 + w * (-69.0 + w * (34.0 + w * (6.0 + w * (-6.0 + w))))))))) / 4320.0
        w2 = w * w
        ws[4] = (78095.0 / 63.0 - w2 * (700.0 + w2 * (-190.0 + w2 * (100.0 / 3.0 + w2 * (-5.0 + w))))) / 2880.0
        ws[5] = (44117.0 / 63.0 + w * (809.0 + w * (44.0 + w * (-868.0 / 3.0 + w * (-86.0 + w * (46.0 + w * (68.0 / 3.0 + w * (-4.0 + w * (-4.0 + w))))))))) / 2880.0
        ws[6] = (3652.0 / 21.0 - w * (-867.0 / 2.0 + w * (-408.0 + w * (-134.0 + w * (48.0 + w * (51.0 + w * (-4.0 + w) * (-1.0 + w) * (2.0 + w))))))) / 4320.0
        ws[7] = (251.0 / 18.0 + w * (123.0 / 2.0 + w * (118.0 + w * (126.0 + w * (77.0 + w * (21.0 + w * (-14.0 / 3.0 + w * (-6.0 + w * (-2.0 + w))))))))) / 10080.0
        ws[9] = w2 * w2
        ws[9] *= ws[9] * w / 362880.0
        ws[8] = 1.0 - ws[0] - ws[1] - ws[2] - ws[3] - ws[4] - ws[5] - ws[6] - ws[7] - ws[9]
        return ws
    ############################################################################################################                                            
    ############################################################################################################                                            
    ############################################################################################################                                                

            
        
        
        
        
        
        
        
        
    """    

        
    
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    public class BetaSpline : Interpolator
            
        public BetaSpline(Single[] bytes, int start, int length, int x, int y, int z, int degree, int copies,bool mirror, bool sample) : base(bytes, start, length, x, y, z,copies)
        {
            TOLERANCE = 2.2204460492503131e-016; // smallest such that 1.0+DBL_EPSILON != 1.0
            _degree = degree;
            _mirror = mirror;
            
            if (_mirror)
            {
                _coefficients = getCubeWhole();
                createCoefficients();
            }
            else if (_sample)
            {
                createSampledCoefficients();
            }
            else
            {
                createPeriodicCoefficients();
            }
        }        
        public override double getValue(double xx, double yy, double zz)
        {            
            if (!isValid(xx, yy, zz))
                return 0;
            double[] xyz = adjustReflection(xx, yy, zz);
            double x = xyz[0];
            double y = xyz[1];
            double z = xyz[2];
                        
            int weight_length = _degree + 1;
            List<int> xIndex = new List<int>();
            List<int> yIndex = new List<int>();
            List<int> zIndex = new List<int>();
            List<double> xWeight = new List<double>();
            List<double> yWeight = new List<double>();
            List<double> zWeight = new List<double>();
            for (int s = 0; s < weight_length; ++s)
            {
                xIndex.Add(0);
                yIndex.Add(0);
                zIndex.Add(0);
                xWeight.Add(0);
                yWeight.Add(0);
                zWeight.Add(0);
            }


            //Compute the interpolation indices
            int i = Convert.ToInt32(Math.Floor(x) - _degree / 2);
            int j = Convert.ToInt32(Math.Floor(y) - _degree / 2);
            int k = Convert.ToInt32(Math.Floor(z) - _degree / 2);

            for (int l = 0; l <= _degree; ++l)
            {
                xIndex[l] = i++; //if 71.1 passed in, for linear, we would want 71 and 72, 
                yIndex[l] = j++;
                zIndex[l] = k++;
            }

            /* compute the interpolation weights */

            if (_degree == 9)
            {
                xWeight = applyValue9(x, xIndex, weight_length);
                yWeight = applyValue9(y, yIndex, weight_length);
                zWeight = applyValue9(z, zIndex, weight_length);
            }
            else if (_degree == 7)
            {
                xWeight = applyValue7(x, xIndex, weight_length);
                yWeight = applyValue7(y, yIndex, weight_length);
                zWeight = applyValue7(z, zIndex, weight_length);
            }
            else if (_degree == 5)
            {
                xWeight = applyValue5(x, xIndex, weight_length);
                yWeight = applyValue5(y, yIndex, weight_length);
                zWeight = applyValue5(z, zIndex, weight_length);
            }
            else
            {
                xWeight = applyValue3(x, xIndex, weight_length);
                yWeight = applyValue3(y, yIndex, weight_length);
                zWeight = applyValue3(z, zIndex, weight_length);
            }

            //applying the mirror boundary condition becaue I am only interpolating within values??
            // RSA edit actually I want to wrap
            int Width2 = 2 * XLen - 2;
            int Height2 = 2 * YLen - 2;
            int Depth2 = 2 * ZLen - 2;
            if (_mirror)
            {
                for (k = 0; k <= _degree; k++)
                {
                    xIndex[k] = (XLen == 1) ? (0) :
                        ((xIndex[k] < 0) ?
                            (-xIndex[k] - Width2 * ((-xIndex[k]) / Width2)) :
                            (xIndex[k] - Width2 * (xIndex[k] / Width2)));
                    if (XLen <= xIndex[k])
                    {
                        xIndex[k] = Width2 - xIndex[k];                    
                    }

                    yIndex[k] = (YLen == 1) ? (0) :
                        ((yIndex[k] < 0) ?
                            (-yIndex[k] - Height2 * ((-yIndex[k]) / Height2)) :
                            (yIndex[k] - Height2 * (yIndex[k] / Height2)));
                    if (YLen <= yIndex[k])
                    {
                        yIndex[k] = Height2 - yIndex[k];                    
                    }

                    zIndex[k] = (ZLen == 1) ? (0) :
                        ((zIndex[k] < 0) ?
                            (-zIndex[k] - Depth2 * ((-zIndex[k]) / Depth2)) :
                            (zIndex[k] - Depth2 * (zIndex[k] / Depth2)));
                    if (ZLen <= zIndex[k])
                    {
                        zIndex[k] = Depth2 - zIndex[k];                    
                    }
                }
            }

            //Perform interolation            
            int splineDegree = _degree;
            double w3 = 0.0;
            for (k = 0; k <= splineDegree; k++)
            {
                double w2 = 0.0;
                for (j = 0; j <= splineDegree; j++)
                {
                    double w1 = 0.0;
                    for (i = 0; i <= splineDegree; i++)
                    {
                        w1 += xWeight[i] * getCoef(xIndex[i], yIndex[j], zIndex[k]);
                    }
                    w2 += yWeight[j] * w1;
                }
                w3 += zWeight[k] * w2;
            }
            return w3;
        }
        private Single[] copyCoef(int buffer, Single[] copyFromCoeffs)
        {
            Single[] copyToCoeffs = new Single[(XLen - (buffer * 2)) * (YLen - (buffer * 2)) * (ZLen - (buffer * 2))];
            int fromCount = 0;
            int toCount = 0;

            for (int k = 0; k < ZLen; k++)
            {                
                for (int j = 0; j < YLen; j++)
                {                    
                    for (int i = 0; i < XLen; i++)
                    {
                        if (i >= buffer && i < XLen - buffer)
                        {
                            if (j >= buffer && j < YLen - buffer)
                            {
                                if (k >= buffer && k < ZLen - buffer)
                                {
                                    copyToCoeffs[toCount] = copyFromCoeffs[fromCount];
                                    ++toCount;
                                }

                            }

                        }
                        ++fromCount;                        
                    }
                }
            }
            return copyToCoeffs;
        }
        private double getCoef(int x, int y, int z)
        {
            int pos = getPosition(x, y, z);
            try
            {
                if (pos >=0 && _coefficients.Length > pos)
                {
                    Single value = _coefficients[pos];
                    return value;
                }
                else
                {
                    return 0;
                }
            }
            catch (Exception e)
            {
                return 0;
            }
            
        }
        private void putCoef(int x, int y, int z, double v)
        {
            int pos = getPosition(x, y, z);
            /*byte[] bv = BitConverter.GetBytes(Convert.ToSingle(v));
            int start = _bStart + pos * 4;
            foreach (byte b in bv)
            {
                _coefficients[start] = b;
                start++;
            }*/
            _coefficients[pos] = Convert.ToSingle(v);
            //if (_coefficients.ContainsKey(pos))
            //    _coefficients[pos] = v;            
        }
        
        private void createSampledCoefficients()
        {
            // we have been given a cube with a buffer, so create as usual then cut
            createCoefficients();
            Single[] copyCoeffs = getCubeEmpty(-1*_buffer);
            copyCoeffs = copyCoef(_buffer, _coefficients);
            _coefficients = copyCoeffs;
            XLen -= (2 * _buffer);
            YLen -= (2 * _buffer);
            ZLen -= (2 * _buffer);
        }
        private void createCoefficients()
        {            
            List<double> pole = getPole(_degree);
            int numPoles = Convert.ToInt32(pole.Count);

            //Convert the samples to interpolation coefficients
            //X-wise
            for (int y = 0; y < YLen; ++y)
            {
                for (int z = 0; z < ZLen; ++z)
                {
                    List<double> row = getRow3d(y, z, XLen);
                    List<double> line = convertToInterpolationCoefficients(pole, numPoles, XLen, row);
                    putRow3d(y, z, line, XLen);
                }
            }
            //Y-wise
            for (int x = 0; x < XLen; ++x)
            {
                for (int z = 0; z < ZLen; ++z)
                {
                    List<double> row = getColumn3d(x, z, YLen);
                    List<double> line = convertToInterpolationCoefficients(pole, numPoles, YLen, row);
                    putColumn3d(x, z, line, YLen);
                }
            }

            //Z-wise
            for (int x = 0; x < XLen; ++x)
            {
                for (int y = 0; y < YLen; ++y)
                {
                    List<double> row = getHole3d(x, y, ZLen);
                    List<double> line = convertToInterpolationCoefficients(pole, numPoles, ZLen, row);
                    putHole3d(x, y, line, ZLen);
                }
            }
        }

        private List<double> getRow3d(int y, int z, int length)
        {
            List<double> row = new List<double>();
            for (int x = 0; x < length; ++x)
                row.Add(getCoef(x, y, z));
            return row;
        }
        private void putRow3d(int y, int z, List<double> row, int length)
        {
            for (int x = 0; x < length; ++x)
                putCoef(x, y, z, row[x]);
        }
        private List<double> getColumn3d(int x, int z, int length)
        {
            List<double> col = new List<double>();
            for (int y = 0; y < length; ++y)
                col.Add(getCoef(x, y, z));
            return col;
        }
        private void putColumn3d(int x, int z, List<double> col, int length)
        {
            for (int y = 0; y < length; ++y)
                putCoef(x, y, z, col[y]);
        }
        private List<double> getHole3d(int x, int y, int length)
        {
            List<double> bore = new List<double>();
            for (int z = 0; z < length; ++z)
                bore.Add(getCoef(x, y, z));
            return bore;
        }
        private void putHole3d(int x, int y, List<double> bore, int length)
        {
            for (int z = 0; z < length; ++z)
                putCoef(x, y, z, bore[z]);
        }

        private List<double> getPole(int degree)
        {
            //Recover the poles from a lookup table #currently only 3 degree, will I want to calculate all the possibilities at the beginnning, 3,5,7,9?
            List<double> pole = new List<double>();
            if (degree == 9)
            {
                pole.Add(-0.60799738916862577900772082395428976943963471853991);
                pole.Add(-0.20175052019315323879606468505597043468089886575747);
                pole.Add(-0.043222608540481752133321142979429688265852380231497);
                pole.Add(-0.0021213069031808184203048965578486234220548560988624);
            }
            else if (degree == 7)
            {
                pole.Add(-0.53528043079643816554240378168164607183392315234269);
                pole.Add(-0.12255461519232669051527226435935734360548654942730);
                pole.Add(-0.0091486948096082769285930216516478534156925639545994);
            }
            else if (degree == 5)
            {
                pole.Add(Math.Sqrt(135.0 / 2.0 - Math.Sqrt(17745.0 / 4.0)) + Math.Sqrt(105.0 / 4.0) - 13.0 / 2.0);
                pole.Add(Math.Sqrt(135.0 / 2.0 + Math.Sqrt(17745.0 / 4.0)) - Math.Sqrt(105.0 / 4.0) - 13.0 / 2.0);
            }
            else//then it is 3
            {
                pole.Add(Math.Sqrt(3.0) - 2.0);
            }
            return pole;
        }

        private List<double> convertToInterpolationCoefficients(List<double> pole, int numPoles, int width, List<double> row)
        {
            /* special case required by mirror boundaries */
            if (width == 1)
            {                
                //mirror filter and periodic filter
                // not much can be done if it is only 1 thivk, it is both mirror and periodic at the same time
                return row; ;                                
            }

            double lambda = 1;
            int n = 0;
            int k = 0;
            //Compute the overall gain
            for (k = 0; k < numPoles; k++)
            {
                lambda = lambda * (1 - pole[k]) * (1 - 1 / pole[k]);
            }
            //Apply the gain
            for (n = 0; n < width; n++)
            {
                row[n] *= lambda;
            }
            //loop over the poles            
            for (k = 0; k < numPoles; k++)
            {
                /* causal initialization */
                if (_mirror)
                {
                    row[0] = InitialCausalCoefficient(row, width, pole[k]);
                }
                else //TODO
                {                    
                    row[0] = InitialCausalCoefficient(row, width, pole[k]);
                }
                /* causal recursion */
                for (n = 1; n < width; n++)
                {
                    row[n] += (double)pole[k] * row[n - 1];
                }
                /* anticausal initialization */
                if (_mirror)
                    row[width - 1] = InitialAntiCausalCoefficient(row, width, pole[k]);
                else //TODO
                    row[width - 1] = InitialAntiCausalCoefficient(row, width, pole[k]);
                /* anticausal recursion */
                for (n = width - 2; 0 <= n; n--)
                {
                    row[n] = pole[k] * (row[n + 1] - row[n]);
                }
            }
            return row;
        }

        private double InitialCausalCoefficient(List<double> c, int dataLength, double pole)

        { /* begin InitialCausalCoefficient */

            double Sum, zn, z2n, iz;
            int n, Horizon;

            /* this initialization corresponds to mirror boundaries */
            Horizon = dataLength;
            if (TOLERANCE > 0.0)
            {
                Horizon = (int)Math.Ceiling(Math.Log(TOLERANCE) / Math.Log(Math.Abs(pole)));
            }
            if (Horizon < dataLength)
            {
                /* accelerated loop */
                zn = pole;
                Sum = c[0];
                for (n = 1; n < Horizon; n++)
                {
                    Sum += zn * c[n];
                    zn *= pole;
                }
                return (Sum);
            }
            else// if (_mirror)
            {// RSA notes this is the mirror condition, when the horizon - how far you need to look ahead for good data - is not as far the data you have
                /* full loop */
                zn = pole;
                iz = 1.0 / pole;
                z2n = Math.Pow(pole, dataLength - 1);
                Sum = c[0] + z2n * c[dataLength - 1];
                z2n *= z2n * iz; //is this a mistake, should it be just *=??? Checked it is how it is in their code. NO TRIED IT.                
                for (n = 1; n <= dataLength - 2; n++)
                {
                    Sum += (zn + z2n) * c[n];
                    zn *= pole;
                    //z2n *= z2n * iz;
                    z2n *= iz;
                }
                return (Sum / (1.0 - zn * zn));
            }
            /*else
            {// RSA notes speculative attempt to make it periodic                
                zn = pole;
                Sum = c[0];
                for (n = 1; n < Horizon; n++)
                {
                    int m = n % dataLength;
                    Sum += zn * c[m];
                    zn *= pole;
                }
                return (Sum);
            }*/
        }
        private double InitialAntiCausalCoefficient(List<double> c, int dataLength, double pole)
        {
            /* this initialization corresponds to mirror boundaries */
            if (dataLength < 2)
                return 0;
            else// if (_mirror)
                return ((pole / (pole * pole - 1.0)) * (pole * c[dataLength - 2] + c[dataLength - 1]));
            //else //wrap to beginning instead
            //    return ((pole / (pole * pole - 1.0)) * (pole * c[1] + c[0]));
        }
        private List<double> applyValue3(double val, List<int> idc, int weight_length)
        {
            List<double> ws = new List<double>();
            for (int i = 0; i < weight_length; ++i)
                ws.Add(0);
            double w = val - (double)idc[1];
            ws[3] = (1.0 / 6.0) * w * w * w;
            ws[0] = (1.0 / 6.0) + (1.0 / 2.0) * w * (w - 1.0) - ws[3];
            ws[2] = w + ws[0] - 2.0 * ws[3];
            ws[1] = 1.0 - ws[0] - ws[2] - ws[3];
            return ws;
        }

        private List<double> applyValue5(double val, List<int> idc, int weight_length)
        {
            List<double> ws = new List<double>();
            for (int i = 0; i < weight_length; ++i)
                ws.Add(0);
            double w = val - (double)idc[2];
            double w2 = w * w;
            ws[5] = (1.0 / 120.0) * w * w2 * w2;
            w2 -= w;
            double w4 = w2 * w2;
            w -= 1.0 / 2.0;
            double t = w2 * (w2 - 3.0);
            ws[0] = (1.0 / 24.0) * (1.0 / 5.0 + w2 + w4) - ws[5];
            double t0 = (1.0 / 24.0) * (w2 * (w2 - 5.0) + 46.0 / 5.0);
            double t1 = (-1.0 / 12.0) * w * (t + 4.0);
            ws[2] = t0 + t1;
            ws[3] = t0 - t1;
            t0 = (1.0 / 16.0) * (9.0 / 5.0 - t);
            t1 = (1.0 / 24.0) * w * (w4 - w2 - 5.0);
            ws[1] = t0 + t1;
            ws[4] = t0 - t1;
            return ws;
        }

        private List<double> applyValue7(double val, List<int> idc, int weight_length)
        {
            List<double> ws = new List<double>();
            for (int i = 0; i < weight_length; ++i)
                ws.Add(0);
            double w = val - (double)idc[3];
            ws[0] = 1.0 - w;
            ws[0] *= ws[0];
            ws[0] *= ws[0] * ws[0];
            ws[0] *= (1.0 - w) / 5040.0;
            double w2 = w * w;
            ws[1] = (120.0 / 7.0 + w * (-56.0 + w * (72.0 + w * (-40.0 + w2 * (12.0 + w * (-6.0 + w)))))) / 720.0;
            ws[2] = (397.0 / 7.0 - w * (245.0 / 3.0 + w * (-15.0 + w * (-95.0 / 3.0 + w * (15.0 + w * (5.0 + w * (-5.0 + w))))))) / 240.0;
            ws[3] = (2416.0 / 35.0 + w2 * (-48.0 + w2 * (16.0 + w2 * (-4.0 + w)))) / 144.0;
            ws[4] = (1191.0 / 35.0 - w * (-49.0 + w * (-9.0 + w * (19.0 + w * (-3.0 + w) * (-3.0 + w2))))) / 144.0;
            ws[5] = (40.0 / 7.0 + w * (56.0 / 3.0 + w * (24.0 + w * (40.0 / 3.0 + w2 * (-4.0 + w * (-2.0 + w)))))) / 240.0;
            ws[7] = w2;
            ws[7] *= ws[7] * ws[7];
            ws[7] *= w / 5040.0;
            ws[6] = 1.0 - ws[0] - ws[1] - ws[2] - ws[3] - ws[4] - ws[5] - ws[7];
            return ws;
        }

        private List<double> applyValue9(double val, List<int> idc, int weight_length)
        {
            List<double> ws = new List<double>();
            for (int i = 0; i < weight_length; ++i)
                ws.Add(0);
            double w = val - (double)idc[4];
            ws[0] = 1.0 - w;
            ws[0] *= ws[0];
            ws[0] *= ws[0];
            ws[0] *= ws[0] * (1.0 - w) / 362880.0;
            ws[1] = (502.0 / 9.0 + w * (-246.0 + w * (472.0 + w * (-504.0 + w * (308.0 + w * (-84.0 + w * (-56.0 / 3.0 + w * (24.0 + w * (-8.0 + w))))))))) / 40320.0;
            ws[2] = (3652.0 / 9.0 - w * (2023.0 / 2.0 + w * (-952.0 + w * (938.0 / 3.0 + w * (112.0 + w * (-119.0 + w * (56.0 / 3.0 + w * (14.0 + w * (-7.0 + w))))))))) / 10080.0;
            ws[3] = (44117.0 / 42.0 + w * (-2427.0 / 2.0 + w * (66.0 + w * (434.0 + w * (-129.0 + w * (-69.0 + w * (34.0 + w * (6.0 + w * (-6.0 + w))))))))) / 4320.0;
            double w2 = w * w;
            ws[4] = (78095.0 / 63.0 - w2 * (700.0 + w2 * (-190.0 + w2 * (100.0 / 3.0 + w2 * (-5.0 + w))))) / 2880.0;
            ws[5] = (44117.0 / 63.0 + w * (809.0 + w * (44.0 + w * (-868.0 / 3.0 + w * (-86.0 + w * (46.0 + w * (68.0 / 3.0 + w * (-4.0 + w * (-4.0 + w))))))))) / 2880.0;
            ws[6] = (3652.0 / 21.0 - w * (-867.0 / 2.0 + w * (-408.0 + w * (-134.0 + w * (48.0 + w * (51.0 + w * (-4.0 + w) * (-1.0 + w) * (2.0 + w))))))) / 4320.0;
            ws[7] = (251.0 / 18.0 + w * (123.0 / 2.0 + w * (118.0 + w * (126.0 + w * (77.0 + w * (21.0 + w * (-14.0 / 3.0 + w * (-6.0 + w * (-2.0 + w))))))))) / 10080.0;
            ws[9] = w2 * w2;
            ws[9] *= ws[9] * w / 362880.0;
            ws[8] = 1.0 - ws[0] - ws[1] - ws[2] - ws[3] - ws[4] - ws[5] - ws[6] - ws[7] - ws[9];
            return ws;
    
    
    
    
    public class OptBSpline : Interpolator
    {
        protected int _degree;
        protected int _dimsize;
        protected int _points;        
        protected int _opt_count = 2;
        protected BetaSpline? _bsp;
        //protected List<Tuple<int[], BetaSpline>> _cubeList = new List<Tuple<int[], BetaSpline>>();
        protected int _xFloor = 0;
        protected int _yFloor = 0;
        protected int _zFloor = 0;
        protected int _xCeil = 0;
        protected int _yCeil = 0;
        protected int _zCeil = 0;

        // ****** Linear Implementation ****************************************
        public OptBSpline(Single[] bytes, int start, int length, int x, int y, int z, int degree, int points, int copies) : base(bytes, start, length, x, y, z,copies)
        {
            // the degree must be an odd number
            _degree = degree;
            _points = points;
            if (_points > XLen)
                _points = XLen;
            if (_points > YLen)
                _points = YLen;
            if (_points > ZLen)
                _points = ZLen;

            _dimsize = (int)Math.Pow(_points, 3);
            _bsp = null;
        }        
        public override double getValue(double xx, double yy, double zz)
        {
            if (!isValid(xx, yy, zz))
                return 0;
            double[] xyz = adjustReflection(xx, yy, zz);
            double x = xyz[0];
            double y = xyz[1];
            double z = xyz[2];

            // The method of linear interpolation is a version of my own method for multivariate fitting, instead of trilinear interpolation
            // NOTE I could extend this to be multivariate not linear but it has no advantage over bspline - and is slower and not as good 
            // Document is here: https://rachelalcraft.github.io/Papers/MultivariateInterpolation/MultivariateInterpolation.pdf

            //1.  do we need a new cube?
            // a) we do if we don't have one

            // b) we do if we are within points/2 of the edges of the edges
            int myxFloor = (int)Math.Floor(x + (-1 * _buffer) + 1);
            int myyFloor = (int)Math.Floor(y + (-1 * _buffer) + 1);
            int myzFloor = (int)Math.Floor(z + (-1 * _buffer) + 1);            
            //int myxFloor = (int)x - (_points / 2 + _buffer);
            //int myyFloor = (int)y - (_points / 2 + _buffer);
            //int myzFloor = (int)z - (_points / 2 + _buffer);

            if (_seedWidth == -1 || _bsp == null)
            {
                selectBSpline(myxFloor, myyFloor, myzFloor, _points);
            }
            else
            {
                checkBSpline(myxFloor, myyFloor, myzFloor);
            }
            

            //2. Find the values that would be in this cube             
            double xn = x - _xFloor;
            double yn = y - _yFloor;
            double zn = z - _zFloor;

            //3. Apply the interpolaor to find the value
            if (_bsp != null)                
                return _bsp.getValue(xn, yn, zn);           
            else
                return 0;
        }

        public void makeCentreBSpline(double x, double y, double z, int points)
        {
            int xFloor = (int)Math.Floor(x + (-1 * _buffer) + 1);
            int yFloor = (int)Math.Floor(y + (-1 * _buffer) + 1);
            int zFloor = (int)Math.Floor(z + (-1 * _buffer) + 1);
            //int xFloor = (int)x - (points / 2 + _buffer);
            //int yFloor = (int)y - (points / 2 + _buffer);
            //int zFloor = (int)z - (points / 2 + _buffer);
            _xFloor = xFloor;
            _yFloor = yFloor;
            _zFloor = zFloor;            
            int buffered_points = points + (_buffer * 2);
            _xCeil = _xFloor + buffered_points;
            _yCeil = _yFloor + buffered_points;
            _zCeil = _zFloor + buffered_points;
            Single[] vals = getSmallerCubeThevenaz(xFloor, yFloor, zFloor, buffered_points);
            //3. Kind of recursive, make a smaller BSlipe interpolator out of this.                             
            _bsp = new BetaSpline(vals, 0, buffered_points * buffered_points * buffered_points, buffered_points, buffered_points, buffered_points, _degree, _copies, true, false);
        }
        private void checkBSpline(int xFloor, int yFloor, int zFloor)
        {
            bool need = false;
            if (!(Math.Abs(xFloor - _xFloor) < _buffer))
                need = true;
            if (!(Math.Abs(yFloor - _yFloor) < _buffer))
                need = true;
            if (!(Math.Abs(zFloor - _zFloor) < _buffer))
                need = true;
        }
        private void selectBSpline(int xFloor, int yFloor, int zFloor, int points)
        {
            bool need = true;            
            //foreach (var cuver in _cubeList)
            //{
            //    int[] points = cuver.Item1;
            need = false;            
            if (!(Math.Abs(xFloor - _xFloor) < _buffer))
                need = true;
            if (!(Math.Abs(yFloor - _yFloor) < _buffer))
                need = true;
            if (!(Math.Abs(zFloor - _zFloor) < _buffer))
                need = true;

            if (!need && _bsp != null)
            {
                //bsp = cuver.Item2;
                //_xFloor = points[0];
                //_yFloor = points[1];
                //_zFloor = points[2];
                //return bsp;
            }
            //}
            else //which it must be or it returned
            {
                _xFloor = xFloor;
                _yFloor = yFloor;
                _zFloor = zFloor;
                int buffered_points = points + (_buffer * 2);
                _xCeil = _xFloor + buffered_points;
                _yCeil = _yFloor + buffered_points;
                _zCeil = _zFloor + buffered_points;
                Single[] vals = getSmallerCubeThevenaz(xFloor, yFloor, zFloor, buffered_points);
                //3. Kind of recursive, make a smaller BSlipe interpolator out of this.                             
                _bsp = new BetaSpline(vals, 0, buffered_points * buffered_points * buffered_points, buffered_points, buffered_points, buffered_points, _degree, 0, true, false);
                //_cubeList.Add(Tuple.Create(new int[]{ xFloor, yFloor, zFloor }, bsp));

                //if (_cubeList.Count > _opt_count)
                //{
                //    int[] vvv = _cubeList[0].Item1;
                //    _cubeList.RemoveAt(0);
                //}
            }            
        }
    }
    
    """


"""


"""
from abc import ABC, abstractmethod
from maptial.xyz import vectorthree as v3
from maptial.xyz import matrix3d as d3
import math
import numpy as np

### Factory method for creation ##############################################################
def create_scoring(method, interpA, interpB,log_level=0):    
    #Factory method to create scoring function classes.
    #Parameters
    #----------
    #method : string
    #    The name of the scoring function method, least squares, abs val...
    #interpA : an interpolator
    #    The interpolator that contains the full matrix of values which is the template
    #interpB : an interpolator
    #    The interpolator that contains the full matrix of values that we are comparing
    
    #log_level : int=0
    #    How much logging you want to see
    #Return
    #------
    #scorer
        
    if log_level > 0:
        print("Scoring:",method)
                                        
    scrr = None
    if method == "absval":
        scrr = AbsValueScore(interpA,interpB,log_level)
    elif method == "leastsquares":
        scrr = LeastSquaresScore(interpA,interpB,log_level)    
    else: 
        raise(Exception("Method not known " + method))
    scrr.init()
    return scrr

### Abstract class ############################################################################
class Scoring(ABC):
    def __init__(self, interpA,interpB,log_level=0):
        self.interpA = interpA
        self.interpB = interpB        
        self.log_level = log_level                 
        self.round = 12        
    ############################################################################################################            
    @abstractmethod
    def get_score(self, central, linear, planar):
        pass    
    ############################################################################################################
    # implemented interface that is the same for all abstractions
    def get_slice3d(self, central,linear,planar):
        pass
        
####################################################################################################
### Absolute Value difference scorer
####################################################################################################
class AbsValueScore(Scoring):                
    def init(self):
        pass
    ## implement abstract interface #########################################
    def get_score(self, central, linear, planar):        
        return 0
####################################################################################################
### Least Squares difference scorer
####################################################################################################
class LeastSquaresScore(Scoring):                
    def init(self):
        pass
    ## implement abstract interface #########################################
    def get_score(self, central, linear, planar):        
        return 0
            
    


"""
https://stackoverflow.com/questions/21836067/interpolate-3d-volume-with-numpy-and-or-scipy

"""
import numpy as np
from numba import njit

@njit(fastmath=True)
def trilinear_interpolation_jit(
    x_volume,
    y_volume,
    z_volume,
    volume,
    x_needed,
    y_needed,
    z_needed
):
    """
    Trilinear interpolation (from Wikipedia)

    :param x_volume: x points of the volume grid 
    :type crack_type: list or numpy.ndarray
    :param y_volume: y points of the volume grid 
    :type crack_type: list or numpy.ndarray
    :param x_volume: z points of the volume grid 
    :type crack_type: list or numpy.ndarray
    :param volume:   volume
    :type crack_type: list or numpy.ndarray
    :param x_needed: desired x coordinate of volume
    :type crack_type: float
    :param y_needed: desired y coordinate of volume
    :type crack_type: float
    :param z_needed: desired z coordinate of volume
    :type crack_type: float

    :return volume_needed: desired value of the volume, i.e. volume(x_needed, y_needed, z_needed)
    :type volume_needed: float
    """

    # dimensinoal check
    assert np.shape(volume) == (
        len(x_volume), len(y_volume), len(z_volume)
    ), "Incompatible lengths"
    # check of the indices needed for the correct control volume definition
    i = np.searchsorted(x_volume, x_needed)
    j = np.searchsorted(y_volume, y_needed)
    k = np.searchsorted(z_volume, z_needed)
    # control volume definition
    control_volume_coordinates = np.array(
        [
            [
                x_volume[i - 1],
                y_volume[j - 1],
                z_volume[k - 1]
            ],
            [
                x_volume[i],
                y_volume[j],
                z_volume[k]
            ]
        ]
    )
    xd = (
        np.array([x_needed, y_needed, z_needed]) - control_volume_coordinates[0]
    ) / (
        control_volume_coordinates[1] - control_volume_coordinates[0]
    )
    # interpolation along x
    c2 = [[0., 0.], [0., 0.]]
    for m, n in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        c2[m][n] = volume[i - 1][j - 1 + m][k - 1 + n] \
        * (1. - xd[0]) + volume[i][j - 1 + m][k - 1 + n] * xd[0]
    # interpolation along y
    c1 = [0., 0.]
    c1[0] = c2[0][0] * (1. - xd[1]) + c2[1][0] * xd[1]
    c1[1] = c2[0][1] * (1. - xd[1]) + c2[1][1] * xd[1]
    # interpolation along z
    volume_needed = c1[0] * (1. - xd[2]) + c1[1] * xd[2]
    return volume_needed

@njit(fastmath=True)
def trilint_jit(
    x_volume,
    y_volume,
    z_volume,
    volume,
    x_needed,
    y_needed,
    z_needed
):
    trilint_size = x_needed.size * y_needed.size * z_needed.size
    jitted_trilint = np.zeros(trilint_size)
    m = 0
    for x in range(0, len(x_needed)):
        for y in range(0, len(y_needed)):
            for z in range(0, len(z_needed)):
                jitted_trilint[m]=trilinear_interpolation_jit(
                    x_volume,
                    y_volume,
                    z_volume,
                    volume,
                    x_needed[x], 
                    y_needed[y],
                    z_needed[z]
                )
                m = m + 1
    return jitted_trilint
############################################
import numpy as np
from itertools import product

def trilinear_interpolation(x_volume, y_volume, z_volume, volume, x_needed, y_needed, z_needed):
    """
    Trilinear interpolation (from Wikipedia)

    :param x_volume: x points of the volume grid 
    :type crack_type: list or numpy.ndarray
    :param y_volume: y points of the volume grid 
    :type crack_type: list or numpy.ndarray
    :param x_volume: z points of the volume grid 
    :type crack_type: list or numpy.ndarray
    :param volume:   volume
    :type crack_type: list or numpy.ndarray
    :param x_needed: desired x coordinate of volume
    :type crack_type: float
    :param y_needed: desired y coordinate of volume
    :type crack_type: float
    :param z_needed: desired z coordinate of volume
    :type crack_type: float

    :return volume_needed: desired value of the volume, i.e. volume(x_needed, y_needed, z_needed)
    :type volume_needed: float
    """
    # dimensinoal check
    if np.shape(volume) != (len(x_volume), len(y_volume), len(z_volume)):
        raise ValueError(f'dimension mismatch, volume must be a ({len(x_volume)}, {len(y_volume)}, {len(z_volume)}) list or numpy.ndarray')
    # check of the indices needed for the correct control volume definition
    i = searchsorted(x_volume, x_needed)
    j = searchsorted(y_volume, y_needed)
    k = searchsorted(z_volume, z_needed)
    # control volume definition
    control_volume_coordinates = np.array(
        [[x_volume[i - 1], y_volume[j - 1], z_volume[k - 1]], [x_volume[i], y_volume[j], z_volume[k]]])
    xd = (np.array([x_needed, y_needed, z_needed]) - control_volume_coordinates[0]) / (control_volume_coordinates[1] - control_volume_coordinates[0])
    # interpolation along x
    c2 = [[0, 0], [0, 0]]
    for m, n in product([0, 1], [0, 1]):
        c2[m][n] = volume[i - 1][j - 1 + m][k - 1 + n] * (1 - xd[0]) + volume[i][j - 1 + m][k - 1 + n] * xd[0]
    # interpolation along y
    c1 = [0, 0]
    c1[0] = c2[0][0] * (1 - xd[1]) + c2[1][0] * xd[1]
    c1[1] = c2[0][1] * (1 - xd[1]) + c2[1][1] * xd[1]
    # interpolation along z
    volume_needed = c1[0] * (1 - xd[2]) + c1[1] * xd[2]
    return volume_needed

def searchsorted(l, x):
    for i in l:
        if i >= x: break
    #return l.index(i)
    return np.where(l == i)


from scipy.interpolate import RegularGridInterpolator
def trilin_interp_regular_grid(x_volume, y_volume, z_volume, volume, x_needed, y_needed, z_needed):
    # dimensinoal check
    if np.shape(volume) != (len(x_volume), len(y_volume), len(z_volume)):
        raise ValueError(f'dimension mismatch, volume must be a ({len(x_volume)}, {len(y_volume)}, {len(z_volume)}) list or numpy.ndarray')
    # trilinear interpolation on a regular grid
    fn = RegularGridInterpolator((x_volume,y_volume,z_volume), volume)
    volume_needed = fn(np.array([x_needed, y_needed, z_needed]))
    return volume_needed
##############################################
import numpy as np
import time

x_volume = np.array([100., 1000.])
y_volume = np.array([0.2, 0.4, 0.6, 0.8, 1])
z_volume = np.array([0, 0.2, 0.5, 0.8, 1.])

the_volume = np.array(
[[[0.902, 0.985, 1.12, 1.267, 1.366],
[0.822, 0.871, 0.959, 1.064, 1.141],
[0.744, 0.77, 0.824, 0.897, 0.954],
[0.669, 0.682, 0.715, 0.765, 0.806],
[0.597, 0.607, 0.631, 0.667, 0.695]],
[[1.059, 1.09, 1.384, 1.682, 1.881],
[0.948, 0.951, 1.079, 1.188, 1.251],
[0.792, 0.832, 0.888, 0.940, 0.971],
[0.726, 0.733, 0.754, 0.777, 0.792],
[0.642, 0.656, 0.675, 0.691, 0.700]]])

x_needed = np.linspace(100, 1000, 10)
y_needed = np.linspace(0.3, 1, 60)
z_needed = np.linspace(0, 1, 7)

start = time.time()
jitted_trilint = trilint_jit(
    x_volume, y_volume, z_volume, the_volume, x_needed, y_needed, z_needed
)
end = time.time()
print('---')
print(f"NUMBA: {end - start}")
print('---')
"""
start = time.time()
manual_trilint = []
for x in x_needed:
    for y in y_needed:
        for z in z_needed:
            manual_trilint.append(
                trilinear_interpolation(
                    x_volume, y_volume, z_volume, the_volume, x, y, z
                )
            )
end = time.time()
print('---')

print(f"Manual: {end - start}")
print('---')
"""
start = time.time()
auto_trilint = []
for x in x_needed:
    for y in y_needed:
        for z in z_needed:
            auto_trilint.append(
                trilin_interp_regular_grid(
                    x_volume, y_volume, z_volume, the_volume, x, y, z
                )
            )
end = time.time()
print('---')
print(f"Auto: {end - start}")
print('---')


'''
# Automaticaly generated InvariantVandermonde matrix
# Generated from leuci-pol interpolation loibrary
# (c) Rachel Alcraft, 2023, Birkbeck College, London University
'''

import numpy as np
class InvariantVandermonde(object):
	def __init__(self):
		self.make_mat()
	def get_invariant(self):
		return self.alcraft
	def make_mat(self):
		self.alcraft = np.zeros((8,8))
		self.alcraft[0,0]=1.0
		self.alcraft[0,1]=0.0
		self.alcraft[0,2]=0.0
		self.alcraft[0,3]=0.0
		self.alcraft[0,4]=0.0
		self.alcraft[0,5]=0.0
		self.alcraft[0,6]=0.0
		self.alcraft[0,7]=0.0
		self.alcraft[1,0]=-1.0
		self.alcraft[1,1]=1.0
		self.alcraft[1,2]=0.0
		self.alcraft[1,3]=0.0
		self.alcraft[1,4]=0.0
		self.alcraft[1,5]=0.0
		self.alcraft[1,6]=0.0
		self.alcraft[1,7]=0.0
		self.alcraft[2,0]=-1.0
		self.alcraft[2,1]=0.0
		self.alcraft[2,2]=1.0
		self.alcraft[2,3]=0.0
		self.alcraft[2,4]=0.0
		self.alcraft[2,5]=0.0
		self.alcraft[2,6]=0.0
		self.alcraft[2,7]=0.0
		self.alcraft[3,0]=1.0
		self.alcraft[3,1]=-1.0
		self.alcraft[3,2]=-1.0
		self.alcraft[3,3]=1.0
		self.alcraft[3,4]=0.0
		self.alcraft[3,5]=0.0
		self.alcraft[3,6]=0.0
		self.alcraft[3,7]=0.0
		self.alcraft[4,0]=-1.0
		self.alcraft[4,1]=0.0
		self.alcraft[4,2]=0.0
		self.alcraft[4,3]=0.0
		self.alcraft[4,4]=1.0
		self.alcraft[4,5]=0.0
		self.alcraft[4,6]=0.0
		self.alcraft[4,7]=0.0
		self.alcraft[5,0]=1.0
		self.alcraft[5,1]=-1.0
		self.alcraft[5,2]=0.0
		self.alcraft[5,3]=0.0
		self.alcraft[5,4]=-1.0
		self.alcraft[5,5]=1.0
		self.alcraft[5,6]=0.0
		self.alcraft[5,7]=0.0
		self.alcraft[6,0]=1.0
		self.alcraft[6,1]=0.0
		self.alcraft[6,2]=-1.0
		self.alcraft[6,3]=0.0
		self.alcraft[6,4]=-1.0
		self.alcraft[6,5]=0.0
		self.alcraft[6,6]=1.0
		self.alcraft[6,7]=0.0
		self.alcraft[7,0]=-1.0
		self.alcraft[7,1]=1.0
		self.alcraft[7,2]=1.0
		self.alcraft[7,3]=-1.0
		self.alcraft[7,4]=1.0
		self.alcraft[7,5]=-1.0
		self.alcraft[7,6]=-1.0
		self.alcraft[7,7]=1.0



"""
RSA 4/2/23


"""
from maptial.xyz import spacetransform as space
from maptial.xyz import crstransform as crs
from maptial.xyz import gridmaker as grid
from maptial.pol import interpolator as pol
from operator import itemgetter
import datetime
import math

#####################################################################
class MapFunctions(object):
    def __init__(self, pdb_code, mobj,pobj, interp_method,as_sd=0,fo=2,fc=-1,log_level=0):
        # PUBLIC INTERFACE
        self.pdb_code = pdb_code        
        self.mobj = mobj
        self.pobj = pobj
        self.interper_vals = [] #default top using only the main density, which for xray is 2Fo-1Fc
        self.as_sd = as_sd
        self.fo=fo
        self.fc=fc
        self.interp_method = interp_method.lower()
        
        if interp_method.lower()=="none":
            self.interper = None
        else:            
            self.make_interper_if_needed(interp_method,log_level,self.fo,self.fc)            
            #self.interper = pol.create_interpolator(interp_method,self.main_interper_vals,self.mobj.F,self.mobj.M,self.mobj.S,log_level=1,as_sd=self.as_sd)                
        self.crs_spc = None
        # data needed
        dim_order =  [int(self.mobj.map_header["01_NC"]),int(self.mobj.map_header["02_NR"]),int(self.mobj.map_header["03_NS"])]
        crs_starts =  [int(self.mobj.map_header["05_NCSTART"]),int(self.mobj.map_header["06_NRSTART"]),int(self.mobj.map_header["07_NSSTART"])]
        axis_sampling =  [int(self.mobj.map_header["08_NX"]),int(self.mobj.map_header["09_NY"]),int(self.mobj.map_header["10_NZ"])]
        map2crs =  [int(self.mobj.map_header["17_MAPC"])-1,int(self.mobj.map_header["18_MAPR"])-1,int(self.mobj.map_header["19_MAPS"])-1]
        cell_dims = [float(self.mobj.map_header["11_X_length"]),float(self.mobj.map_header["12_Y_length"]),float(self.mobj.map_header["13_Z_length"])]
        angles =  [float(self.mobj.map_header["14_Alpha"]),float(self.mobj.map_header["15_Beta"]),float(self.mobj.map_header["16_Gamma"])]
        self.crs_spc = crs.CrsTransform(dim_order, crs_starts, axis_sampling, map2crs, cell_dims, angles)        
             
    def make_interper_if_needed(self,interp_method,log_level,fo,fc):        
        vs,ds = 1,0
        calc_fofc = False
        create_interp = False
        if self.interper_vals == []:
            create_interp = True
            #if self.mobj.diff_values == [] or fo==2 and fc == -1:
            if self.mobj.diff_has == 0 or fo==2 and fc == -1:
                for i in range(len(self.mobj.values)):
                    self.interper_vals.append(self.mobj.values[i])
            else:
                calc_fofc = True
                
        elif self.mobj.diff_has != 0:
        #elif self.mobj.diff_values == []:
            if (self.interp_method != interp_method or self.fo != fo or self.fc != fc):
                calc_fofc = True
                create_interp = True
        if calc_fofc:
            self.interper_vals = []
            vs, ds = 0,0
            vs, ds = fo, -1 * fo                
            vs = vs + fc
            ds = ds + (-2 * fc)
            
            for i in range(len(self.mobj.values)):
                self.interper_vals.append(vs*self.mobj.values[i] + ds*self.mobj.diff_values[i])
        
        if create_interp:
            if log_level > 0:
                print("New interper, Fos=",fo,"Fcs=",fc,"mains=",vs,"diffs=",ds)
            self.interper = pol.create_interpolator(interp_method,self.interper_vals,(self.mobj.F,self.mobj.M,self.mobj.S),log_level=log_level,as_sd=self.as_sd)
        self.interp_method = interp_method
        self.fo = fo
        self.fc = fc
                    
    def max_min(self):
        return self.interper.min,self.interper.max

    def get_slices(self,central, linear, planar, width, samples, interp_method, derivs=[0], fo=2,fc=-1,log_level=0,degree=-1,ret_type="np"):
        vals = []
        for deriv in derivs:
            val = self.get_slice(central, linear, planar, width, samples, interp_method, deriv, fo,fc,log_level,degree,ret_type=ret_type)
            vals.append(val)
        return vals

    def get_slice(self,central, linear, planar, width, samples, interp_method, depth_samples=1, deriv=0, fo=2,fc=-1,log_level=0, ret_type="np"):
        # change interpolator if necessary        
        self.make_interper_if_needed(interp_method,log_level,fo,fc)        
        #############        
        # objects needed
        spc = space.SpaceTransform(central, linear, planar)        
        gm = grid.GridMaker()        
        #########                                        
        u_coords = gm.get_unit_grid(width,samples,depth_samples=depth_samples)
        xyz_coords = spc.convert_coords(u_coords)
        crs_coords = self.crs_spc.convert_coords_to_crs(xyz_coords)
        if depth_samples > 1:    
            vals = self.interper.get_val_slice3d(crs_coords,deriv=deriv,ret_type=ret_type)        
            return vals,xyz_coords
        else:
            vals = self.interper.get_val_slice(crs_coords,deriv=deriv,ret_type=ret_type)        
            return vals
    
    def get_slice_neighbours(self,central, linear, planar, width, samples,rnge,log_level=0):
        #############        
        # objects needed
        spc = space.SpaceTransform(central, linear, planar)
        gm = grid.GridMaker()        
        #########                                        
        u_coords = gm.get_unit_grid(width,samples)                
        xyz_coords = spc.convert_coords(u_coords)                
        in_scope = self.pobj.get_inscope_atoms(central,width+rnge[1],log_level)                
        a,b,c = xyz_coords.shape()
        naybs = []
        for i in range(a):
            row = []
            for j in range(b):
                coord = xyz_coords.get(i,j)                
                nayb = self.pobj.get_neighbours(coord,rnge,in_scope)
                row.append(nayb)
            naybs.append(row)
        return naybs
            
    def get_xyz(self,crs_coords):
        xyz_coords = self.crs_spc.crs_to_xyz(crs_coords)        
        return xyz_coords
    
    def get_crs(self,xyz_coords):
        crs_coords = self.crs_spc.xyz_to_crs(xyz_coords)        
        return crs_coords

    def get_map_projection(self, sliced, xmin=-1,xmax=-1,ymin=-1,ymax=-1):
        vals = self.interper.get_projection(sliced.lower(), xmin,xmax,ymin,ymax)        
        return vals
        
    def get_atoms_projection(self, interp_method,log_level=0):
        
        minx,miny,minz = 1000,1000,1000
        maxx,maxy,maxz = -1000,-1000,-1000

        atoms = self.pobj.get_atom_coords()
        self.make_interper_if_needed(interp_method,log_level,2,-1)
        crss = []        
        for atm in atoms:
            crs_coord = self.get_crs(atm)
            val = self.interper.get_value(crs_coord.A,crs_coord.B,crs_coord.C)
            crss.append((crs_coord,val))
        # now sort on the values
        sorted_crs = sorted(crss,key=itemgetter(1))
        xs,ys,zs,vs = [],[],[],[]
        for co,va in sorted_crs:
            xs.append(co.A)
            ys.append(co.B)
            zs.append(co.C)
            vs.append(va)
                        
            minx = math.floor(min(minx,float(co.A)))
            maxx = math.ceil(max(maxx,float(co.A)))
            miny = math.floor(min(miny,float(co.B)))
            maxy = math.ceil(max(maxy,float(co.B)))
            minz = math.floor(min(minz,float(co.C)))
            maxz = math.ceil(max(maxz,float(co.C)))                     
        
        return xs,ys,zs,vs,(minx,maxx),(miny,maxy),(minz,maxz)
    
    def get_map_cross_section(self, sliced,layer):
        vals = self.interper.get_cross_section(sliced.lower(),layer)        
        return vals

        



        
        
        
        


"""
RSA 4/2/23


"""

import os
from os.path import exists
import urllib.request
from Bio.PDB.MMCIFParser import MMCIFParser        
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB.PDBParser import PDBParser

import struct
import numpy as np

import warnings
from Bio import BiopythonWarning
warnings.simplefilter('ignore', BiopythonWarning)

from . import mapobject as mobj
from . import pdbobject as pobj
from . import mapfunctions as mfun

class MapLoader(object):
    def __init__(self, pdb_code, directory="", cif=False):
        # PUBLIC INTERFACE
        self.mobj = mobj.MapObject(pdb_code)
        self.pobj = pobj.PdbObject(pdb_code)
        
        # Private data
        self._ccp4_binary = None
        self._diff_binary = None        
        self.pdb_loaded = False        
        self.em_loaded = False       
        self.values_loaded = False
        self.values_loading = False
        self.has_both = True
        self.mfunc = None
        self.mean = 0
        self.std = 1
        self.mean_diff = 0
        self.std_diff = 1
        # PRIVATE INTERFACE
        self._directory = directory        
        self._cif=cif
        if cif:
            self._filepath = f"{directory}{pdb_code}.cif"
            self.mobj.pdb_link = f"https://www.ebi.ac.uk/pdbe/entry-files/download/{pdb_code}.cif"
        else:        
            self._filepath = f"{directory}{pdb_code}.pdb"
            self.mobj.pdb_link = f"https://www.ebi.ac.uk/pdbe/entry-files/download/pdb{pdb_code}.ent"
        
        self._filepath_ccp4 = f"{self._directory}{self.mobj.pdb_code}.ccp4"
        self._filepath_diff = f"{self._directory}{self.mobj.pdb_code}_diff.ccp4"
                                        
    def exists(self):
        if self.exists_pdb():
            if self.exists_map():
                return True
        
        return False
    
    def exists_pdb(self):
        if exists(self._filepath) and exists(self._filepath+".done.txt"):
            return True
        else:
            return False
    
    def exists_map(self):
        self.load_pdb()
        if 'x-ray' in self.mobj.exp_method:
            if exists(self._filepath_ccp4) and exists(self._filepath_diff) and exists(self._filepath_ccp4+".done.txt"):
                #self.em_loaded = True
                return True        
            else:
                return False
        elif 'electron' in self.mobj.exp_method.lower():
            if exists(self._filepath_ccp4) and exists(self._filepath_ccp4+".done.txt"):
                #self.em_loaded = True
                return True
            else:
                return False
        else:
            #self.em_loaded = True
            return True # it doesn;t NOT exists anyway
    
    def success(self):
        if exists(self._filepath) and exists(self._filepath_ccp4):
            return True
        else:
            return False

    def download(self):                
        if not self.exists_pdb():
            if not self.already_started_pdb():
                self.write_starting_pdb()
                self.download_pdb()
        if not self.exists_map():
            if not self.already_started_map():
                self.write_starting_map()
                self.download_map()
        self.write_done()
    
    def already_started_pdb(self):
        return exists(self._filepath+".start.txt")    
    def already_started_map(self):
        return exists(self._filepath_ccp4+".start.txt")
    def write_starting_pdb(self):
        with open(self._filepath + ".start.txt","w") as fs_pdb:
            fs_pdb.write("start")            
    def write_starting_map(self):        
        with open(self._filepath_ccp4 + ".start.txt","w") as fs_map:
            fs_map.write("start")
    def write_done(self):
        if exists(self._filepath+".start.txt"):
            os.remove(self._filepath+".start.txt")
        if exists(self._filepath_ccp4+".start.txt"):
            os.remove(self._filepath_ccp4+".start.txt")
        if exists(self._filepath):
            with open(self._filepath + ".done.txt","w") as fw_pdb:
                fw_pdb.write("done")
        else:
            with open(self._filepath + ".failed.txt","w") as fw_pdb:
                fw_pdb.write("failed")
        if exists(self._filepath_ccp4):
            with open(self._filepath_ccp4 + ".done.txt","w") as fw_map:
                fw_map.write("done")
        else:
            with open(self._filepath_ccp4 + ".failed.txt","w") as fw_map:
                fw_map.write("failed")
                
    def download_pdb(self):
        self._fetch_pdbdata()

    def download_map(self):
        if not self.pdb_loaded:
            self.load_pdb()            
        if 'x-ray' in self.mobj.exp_method:
            self._fetch_maplink_xray()            
        elif 'electron' in self.mobj.exp_method:            
            print("Now downloading em matrix")
            self._fetch_maplink_em()
        #self.em_loaded = True
    
    def load(self):
        self.pdb_loaded = self.load_pdb()
        #if 'x-ray' in self.mobj.exp_method:            
        self.em_loaded = self.load_map()
            
            
        

    def load_pdb(self):
        if self._cif:
            structure = MMCIFParser().get_structure(self.mobj.pdb_code, self._filepath)
        else:
            structure = PDBParser(PERMISSIVE=True).get_structure(self.mobj.pdb_code, self._filepath)                
            found_em = False
            with open(self._filepath,"r") as fr:
                lines = fr.readlines()
                for line in lines:
                    self.pobj.add_line_string(line)
                    #REMARK 900 RELATED ID: EMD-6240   RELATED DB: EMDB                              
                    if "REMARK 900 RELATED ID:" in line and "EMD-" in line and not found_em:
                        found_em=True
                        info = line.split(" ")
                        self.mobj.em_code = info[4]
                        self.em_code = self.mobj.em_code
                        self.has_both = False
                        self.mobj.em_link = f"https://www.ebi.ac.uk/emdb/{self.em_code}"
                
        self._struc_dict = MMCIF2Dict(self._filepath)
        self.mobj.resolution = structure.header["resolution"]
        self.mobj.exp_method = structure.header["structure_method"]

        # with the structure create a pdbobject
        return True

    def load_map(self):        
        try:
            with open(self._filepath_ccp4, mode='rb') as file:
                self._ccp4_binary = file.read()
            if self.has_both:
                with open(self._filepath_diff, mode='rb') as file:
                    self._diff_binary = file.read()
            
            self._create_mapheader(self._ccp4_binary)            
            
            return True
        except Exception as e:
            print("Error loading map", str(e))
            return False
    
    def wait_for_load(self,log_level=0):
        import time
        count = 0
        while self.values_loading:
            time.sleep(0.1)
            if (log_level > -0):
                count += 1
                if count%30 == 0:
                    print("Waiting for load values", len(self.mobj.values), len(self.mobj.diff_values))
        return self.values_loaded

    def load_values(self, diff=True, adj_zero=True):
        try:            
            self.values_loading = True
            self._create_mapvalues(False,adj_zero=adj_zero)
            if diff and self.has_both:
                self._create_mapvalues(True,adj_zero=adj_zero)
            if not self.has_both:
                self.mobj.diff_values = []
            self.values_loading = False
            self.values_loaded = True
        except Exception as e:
            print(str(e))
            self.values_loaded = False
            self.values_loading = False


    def get_or_make_func(self, interp):
        if self.mfunc == None:
            self.mfunc = mfun.MapFunctions(self.mobj.pdb_code,self.mobj,self.pobj, interp) #the default method is linear
        return self.mfunc
    #################################################
    ############ PRIVATE INTERFACE ##################
    #################################################
    def _fetch_pdbdata(self):
        try:
            print(self.mobj.pdb_link, self._filepath)            
            urllib.request.urlretrieve(self.mobj.pdb_link, self._filepath)                                
        except:            
            return False
        return True
        
    def _fetch_maplink_xray(self):
        try:
            if not exists(self._filepath_ccp4):            
                urllib.request.urlretrieve(self.mobj.ccp4_link, self._filepath_ccp4)
            if not exists(self._filepath_diff):            
                urllib.request.urlretrieve(self.mobj.diff_link, self._filepath_diff)
            self.mobj.em_code = self.mobj.pdb_code
            self.em_code = self.mobj.pdb_code
        except Exception as e:
            print(e)

                        
    def _fetch_maplink_em(self):     
        """
        cif file
        EMDB EMD-6240 'associated EM volume' . 
        pdb file
        REMARK 900 RELATED ID: EMD-6240   RELATED DB: EMDB                              
        """
        self.mobj.em_code = self.em_code.split("-")[1]
        self.mobj.diff_link = "" # there is no difference density for cryo-em   
        self.mobj.ccp4_link = "https://ftp.ebi.ac.uk/pub/databases/emdb/structures/EMD-" + self.mobj.em_code + "/map/emd_" + self.mobj.em_code + ".map.gz"        
        self.mobj.em_link = f"https://www.ebi.ac.uk/emdb/EMD-{self.em_code}"
        self.has_both = False
        # Download the files        
        filepath_gz = self._filepath_ccp4 + ".gz"
        if not exists(self._filepath_ccp4):
            # need to unzip file
            print("getting",self._filepath_ccp4)
            if not exists(filepath_gz):
                # need to download zipped file
                print("getting",self.mobj.ccp4_link,"to",filepath_gz)
                try:
                    urllib.request.urlretrieve(self.mobj.ccp4_link, filepath_gz)
                except Exception as e:
                    print(str(e))
            # Unzipping
            import gzip
            import shutil
            with gzip.open(filepath_gz, 'rb') as f_in:
                with open(self._filepath_ccp4, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)
                
        

    def _create_mapheader(self, ccp4_binary):
        num_labels = 0
        num_sym = 0        
        headers = [] #https://www.ccp4.ac.uk/html/maplib.html#description
        xheaders = [] 
        self.mobj.header_as_string = ""
        headers.append(["01_NC","int",4])           # of Columns    (fastest changing in map)
        headers.append(["02_NR","int",4])           # of Rows
        headers.append(["03_NS","int",4])           # of Sections   (slowest changing in map)        
        headers.append(["04_MODE","int",4])         # Data type   0 = signed bytes (from-128 lowest to 127 highest) 1 = Integer*2 2 = Image stored as Reals 3 = Complex Integer*2 4 = Complex Reals 5 == 0
        headers.append(["05_NCSTART","int",4])      # Number of first COLUMN  in map
        headers.append(["06_NRSTART","int",4])      # Number of first ROW     in map
        headers.append(["07_NSSTART","int",4])      # Number of first SECTION in map
        headers.append(["08_NX","int",4])           # Number of intervals along X
        headers.append(["09_NY","int",4])           # Number of intervals along Y
        headers.append(["10_NZ","int",4])           # Number of intervals along Z
        headers.append(["11_X_length","double",4])  # Cell Dimensions (Angstroms)
        headers.append(["12_Y_length","double",4])  #             "
        headers.append(["13_Z_length","double",4])  #             "
        headers.append(["14_Alpha","double",4])     # Cell Angles     (Degrees)
        headers.append(["15_Beta","double",4])      #             "
        headers.append(["16_Gamma","double",4])     #             "
        headers.append(["17_MAPC","int",4])         # Which axis corresponds to Cols.  (1,2,3 for X,Y,Z)
        headers.append(["18_MAPR","int",4])         # Which axis corresponds to Rows   (1,2,3 for X,Y,Z)
        headers.append(["19_MAPS","int",4])         # Which axis corresponds to Sects. (1,2,3 for X,Y,Z)
        headers.append(["20_AMIN","double",4])      # Minimum density value
        headers.append(["21_AMAX","double",4])      # Maximum density value
        headers.append(["22_AMEAN","double",4])     # Mean    density value    (Average)
        headers.append(["23_ISPG","int",4])         # Space group number
        headers.append(["24_NSYMBT","int",4])       # Number of bytes used for storing symmetry operators
        headers.append(["25_LSKFLG","int",4])       # Flag for skew transformation, =0 none, =1 if foll
        for i in range(26,35):
            headers.append([str(i) + "_SKWMAT","double",4])       # Flag for skew transformation, =0 none, =1 if foll
        for i in range(35,38):
            headers.append([str(i) + "_SKWTRN","double",4])       # Flag for skew transformation, =0 none, =1 if foll
        for i in range(38,53):
            headers.append(["X","int",4])       # Flag for skew transformation, =0 none, =1 if foll
        headers.append(["53_MAP","string",4])       # Character string 'MAP ' to identify file type
        headers.append(["54_MACHST","int",4])       # Machine stamp indicating the machine type
        headers.append(["55_ARMS","double",4])       # Rms deviation of map from mean density
        headers.append(["56_NLABL","int",4])       # Number of labels being used
                
        i=0
        for header, typ,inc  in headers:
            val = ""
            if not header == "X":
                if typ == "int":
                    val = int.from_bytes(ccp4_binary[i:i+inc], byteorder='little', signed=True)
                    self.mobj.map_header[header] = val
                elif typ == "double":
                    val = struct.unpack('f', ccp4_binary[i:i+inc])[0]
                    self.mobj.map_header[header] = val
                elif typ == "string":
                    val = ccp4_binary[i:i+inc].decode("utf-8") 
                    self.mobj.map_header[header] = val
                    
                if len(header) > 7:
                    self.mobj.header_as_string += header + "\t" + str(val) + "\n"
                else:
                    self.mobj.header_as_string += header + "\t\t" + str(val) + "\n"
                
                if header == "24_NSYMBT":
                    num_sym = int(val/80)                              
                if header == "56_NLABL":
                    num_labels = int(val)
                                                                
            i+=inc
                
        for s in range(0,num_labels):
            xheaders.append([str(s+1) + "_LABEL","string",80])       # 10  80 character text labels (ie. A4 format)
        for s in range(num_labels,10):
            xheaders.append(["X","string",80])       # 10  80 character text labels (ie. A4 format)            
        for s in range(0,num_sym):
            xheaders.append([str(s+1) + "_SYM","string",80])       # 10  80 character text labels (ie. A4 format)                        
            
        
        for header, typ,inc  in xheaders:
            if not header == "X":
                if typ == "int":
                    val = int.from_bytes(ccp4_binary[i:i+inc], byteorder='little', signed=True)
                    self.mobj.map_header[header] = val
                elif typ == "double":
                    val = struct.unpack('f', ccp4_binary[i:i+inc])[0]
                    self.mobj.map_header[header] = val
                elif typ == "string":
                    val = ccp4_binary[i:i+inc].decode("utf-8") 
                    self.mobj.map_header[header] = val
                    
                if len(header) > 7:
                    self.mobj.header_as_string += header + "\t" + str(val) + "\n"
                else:
                    self.mobj.header_as_string += header + "\t\t" + str(val) + "\n"
            i+=inc
        
                        
    def _create_mapvalues(self, diff, adj_zero=True):
        use_binary = self._ccp4_binary
        vals = []
        #vals = {} #zeros alternative
        if diff:
            use_binary = self._diff_binary
        
        Blength = self.mobj.map_header["01_NC"] * self.mobj.map_header["02_NR"] * self.mobj.map_header["03_NS"]
        Bstart = len(self._ccp4_binary) - (4 * Blength)

        self.mobj.F = self.mobj.map_header["01_NC"]
        self.mobj.M = self.mobj.map_header["02_NR"]
        self.mobj.S = self.mobj.map_header["03_NS"]

        if diff:
            self.mobj.diff_values = np.zeros((self.mobj.F,self.mobj.M,self.mobj.S))
            self.mobj.diff_has = self.mobj.F*self.mobj.M*self.mobj.S
        else:
            self.mobj.values = np.zeros((self.mobj.F,self.mobj.M,self.mobj.S))

        count = 0
        for s in range(0,self.mobj.S): #slow
            for m in range(0,self.mobj.M):#medium
                for f in range(0,self.mobj.F):#fast
                    strt = Bstart+(count*4)
                    val = struct.unpack('f', use_binary[strt:strt+4])[0]
                    count += 1
                    if diff:
                        self.mobj.diff_values[f,m,s] = val
                    else:
                        self.mobj.values[f,m,s] = val
                            
        #if adj_zero:
        #    if diff:
        #        zro = 
        #        self.mobj.diff_values = np.zeros((self.mobj.F,self.mobj.M,self.mobj.S))
        #    else:
        #        self.mobj.values = np.zeros((self.mobj.F,self.mobj.M,self.mobj.S))

        """
        Blength = self.mobj.map_header["01_NC"] * self.mobj.map_header["02_NR"] * self.mobj.map_header["03_NS"]
        Bstart = len(self._ccp4_binary) - (4 * Blength)
        for i in range(0,Blength):
            strt = Bstart+(i*4)
            val = struct.unpack('f', use_binary[strt:strt+4])[0]
            vals.append(val)
            #if abs(val) > 0.001:                
            #    vals[i] = val
        if diff:
            self.mobj.diff_values = vals
        else:
            self.mobj.values = vals
        """



    def _create_mapdata_em(self):
        ccp4_link = ""
        em_link = ""


    def cleanup(self):
        pass



#####################################################################
#RSA 17/03/23
#This helper function makes it easier to plot for demonstrations and examples, paritclarly on the colab page
#####################################################################

from maptial.xyz import spacetransform as space
import plotly.graph_objs as go
from plotly.subplots import make_subplots
import matplotlib

#####################################################################
class MapPlotHelp(object):
    def __init__(self,filename):
        # PUBLIC INTERFACE        
        self.filename = filename                
        self.plots_2d = []
        self.plots_3d = []                
            
    def add_plot_slice_2d(self,vals,tag):
        self.plots_2d.append((vals,tag))
    
    def add_plot_slice_3d(self,vals,tag):
        self.plots_3d.append((vals,tag))
                    
    def make_plot_slice_3d(self,vals,min_percent=1, max_percent=1,
                            hue="GBR",title="Leucippus Plot 3d",levels=20,
                            transparency="medium"):
        
        #Takes a mat3d object and plots it in plotly

        #Input
        #---------
        #vals : mat3d
        #min_percent : 1
        #max_percent : 1
        #hue : GBR/GRB/BW/WB/RB/BR/GI/IG/MASK
        #title : "Leucippud Plot 3d
        #levels : 20        
        #https://plotly.com/python/3d-isosurface-plots/
        #turn data into scatter for iso_surface
        
        xs = []
        ys = []
        zs = []
        values = []

        #fig = make_subplots(rows=1, cols=1,subplot_titles=[title],horizontal_spacing=0.05,vertical_spacing=0.05)
        
        minv = 1000
        maxv = -1000

        a,b,c = vals.shape()
        for i in range(a):
            for j in range(b):
                for k in range(a):
                    val = 0
                    if k < c:
                        val = vals.get(i,j,k=k)
                        minv = min(minv,val)
                        maxv = max(maxv,val)
                    xs.append(i)
                    ys.append(j)
                    zs.append(k)
                    values.append(val)
                                    
        
        #if hue == "RINGS":
        #    minv,maxv =  -0.5,1.5
        #    min_percent,max_percent = 1,1
        
        #if hue != "MASK":
        absmin,absmax,d0,d1,d2 = self.__get_levels__(min_percent,max_percent,minv,maxv,hue=="MASK")
        #else:
        #absmin,absmax,d0,d1,d2 = minv,maxv,1,1,1
        
        
        
        colorscale=self.__get_colors__(hue,d0,d1,d2,min_percent,max_percent,transparency=transparency)
                
        fig= go.Figure(data=go.Isosurface(
        x=xs,
        y=ys,
        z=zs,
        value=values,        
        colorscale=colorscale,
        showscale=True,
        showlegend=False,        
        opacity=0.6,
        surface_count=levels,
        caps=dict(x_show=False, y_show=False,z_show=False),
        isomin=absmin,
        isomax=absmax,
        ),)
        
        fig.update_xaxes(showticklabels=False,visible=False,scaleanchor="x",scaleratio=1) # hide all the xticks
        fig.update_yaxes(showticklabels=False,visible=False,scaleanchor="x",scaleratio=1) # hide all the yticks
                                
        fig.update_layout(title=dict(text=title),
                            scene = dict(
                            xaxis = dict(visible=False),
                            yaxis = dict(visible=False),
                            zaxis =dict(visible=False)
                        )
    )

                        
        #print(values)
        if self.filename == "SHOW":
            fig.show()
        elif self.filename == "FIG":
            return fig
        elif ".html" in self.filename:
            fig.write_html(self.filename)
        else:
            fig.write_image(self.filename,width=2000,height=2000)

    def make_plot_slice_2d(self,vals2d,plot_type="contour",points=[], 
                            naybs=[],min_percent=1, max_percent=1,
                            hue="GBR",levels=20,title="Leucippus Plot 2d",samples=-1,width=-1,
                            transparency="no"):
        
        #Takes a mat2d object and plots it in plotly

        #Input
        #---------
        #vals : mat2d
        #plottype : "countour" or heatmap
        #points : [] a selection of points to add ontop of plot
        #naybs : [] neighbours matched per value points
        #min_percent : 1
        #max_percent : 1
        #hue : GBR/GRB/BW/WB/RB/BR/GI/IG/MASK
        #title : "Leucippud Plot 2d
        #levels : 20
        #samples :-1 needed to place the points in the right place
        #width : -1 needed to place the points in the right place
        
        #https://plotly.com/python/3d-isosurface-plots/
        #turn data into scatter for iso_surface        
        
        vals = vals2d.tolist()                
        fig = make_subplots(rows=1, cols=1,subplot_titles=[title],horizontal_spacing=0.05,vertical_spacing=0.05)

        mind,maxd = 1000,-1000    
        for i in range(len(vals)):
            for j in range(len(vals[0])):        
                mind = min(vals[i][j],mind)
                maxd = max(vals[i][j],maxd)        
        
        #if hue == "RINGS":
        #    mind,maxd = -0.5,1.5
        #    min_percent,max_percent = 1,1
        
        #if hue != "MASK":
        absmin,absmax,d0,d1,d2 = self.__get_levels__(min_percent,max_percent,mind,maxd,hue=="MASK")        
        #else:
        #    absmin,absmax,d0,d1,d2 = mind,maxd,1,1,1
        colorscale=self.__get_colors__(hue,d0,d1,d2,min_percent,max_percent,transparency=transparency,)

        if len(naybs) > 0:
            if plot_type == "contour":
                data_vals = go.Contour(z=vals,showscale=False, 
                                colorscale=colorscale,
                                contours=dict(start=absmin,end=absmax,size=(absmax-absmin)/levels),
                                text=naybs,   
                                hovertemplate='......%{z:.4f}<br>%{text}',
                                line=dict(width=0.5,color="gray"),
                                zmin=absmin,zmax=absmax,name='')
            elif plot_type == "heatmap":
                data_vals = go.Heatmap(z=vals,showscale=False, 
                                colorscale=colorscale,                                
                                text=naybs,   
                                hovertemplate='......%{z:.4f}<br>%{text}',                                
                                zmin=absmin,zmax=absmax,name='')
        else:
            if plot_type == "contour":
                data_vals = go.Contour(z=vals,showscale=True, 
                                colorscale=colorscale,
                                contours=dict(start=absmin,end=absmax,size=(absmax-absmin)/levels),
                                hovertemplate='......%{z:.4f}',
                                line=dict(width=0.5,color="gray"),
                                zmin=absmin,zmax=absmax,name='')
            elif plot_type == "heatmap":
                data_vals = go.Heatmap(z=vals,showscale=True, 
                                colorscale=colorscale,                                
                                hovertemplate='......%{z:.4f}',                                
                                zmin=absmin,zmax=absmax,name='')

                        
        fig.add_trace(data_vals,row=1,col=1)
        if len(points) == 3:
            data_scatter = self.add_points(points,samples,width)
            fig.add_trace(data_scatter,row=1,col=1)    
        fig.update_xaxes(showticklabels=False,visible=False) # hide all the xticks
        fig.update_yaxes(showticklabels=False,visible=False) # hide all the xticks
        fig.update_yaxes(scaleanchor="x",scaleratio=1)    
        fig.update_xaxes(scaleanchor="y",scaleratio=1)
                        
        #print(values)
        if self.filename == "SHOW":
            fig.show()
        elif self.filename == "FIG":
            return fig
        elif ".html" in self.filename:
            fig.write_html(self.filename)        
        else:
            fig.write_image(self.filename,width=2000,height=2000)

    def add_points(self, points,samples,width,log_level=0):
        # First create the dots for the potitions as a scatter plot
        spc = space.SpaceTransform(points[0], points[1], points[2])
        posC = spc.reverse_transformation(points[0])
        posL = spc.reverse_transformation(points[1])
        posP = spc.reverse_transformation(points[2])
        posCp = posC.get_point_pos(samples,width)
        posLp = posL.get_point_pos(samples,width)
        posPp = posP.get_point_pos(samples,width)
        if log_level > 0:
            print("Central scatter=",posCp.get_key())
            print("Linear scatter=",posLp.get_key())
            print("Planar scatter=",posPp.get_key())
        scatterX = []
        scatterY = []
        # The C value will be zero as it is on the plane - that is because these are the points we made the plane with
        # The xy heatmap has been arranged so the x value is above so linear is upwards, so the y axis (ok a bit confusing.... should I change it)?
        if posCp.A > 0 and posCp.A < samples and posCp.B > 0 and posCp.B < samples:
            scatterX.append(posCp.B)
            scatterY.append(posCp.A)
        if posLp.A > 0 and posLp.A < samples and posLp.B > 0 and posLp.B < samples:
            scatterX.append(posLp.B)
            scatterY.append(posLp.A)
        if posPp.A > 0 and posPp.A < samples and posPp.B > 0 and posPp.B < samples:
            scatterX.append(posPp.B)
            scatterY.append(posPp.A)  
  ###############################################################################        
                    
        data_scatter = go.Scatter(x=scatterX,y=scatterY,mode="markers",marker=dict(color="yellow",size=5),showlegend=False,hoverinfo='skip',hovertemplate='',name='')            

        return data_scatter

    def __get_levels__(self,min_rr, max_rr, min_vv, max_vv,absolute):        
        
        if absolute:
            mn = min(min_rr,max_rr)
            mx = max(min_rr,max_rr)
            absmin = (mn*(max_vv-min_vv)) + min_vv
            absmax = (mx*(max_vv-min_vv)) + min_vv
        else:
            absmin = min_vv*min_rr        
            absmax = max_vv*max_rr          
            
        if absmin == absmax:
            d0 = 0.5
        elif absmin > 0:
            d0 = 0
        else:
            d0 = (0 - absmin) / (absmax - absmin)                            
        d1 = d0 + ((1-d0)/3)
        d2 = d0 + (2*(1-d0)/3)                        
        return absmin,absmax,d0,d1,d2
    
    def __get_colors__(self,hue,d0,d1,d2,min_r,max_r,transparency):
        t0,t1,t2,t3,t4,tx,tl,tm,tu = 1,1,1,1,1,1,1,1,1
        if transparency == "low":
            t0,t1,t2,t3,t4 = 0.5,0.5,0.5,0.5,0.5
            tx = 0.5
            tl,tm,tu = 0,0.5,1
        elif transparency == "medium":
            t0,t1,t2,t3,t4 = 0.4,0.1,0.4,0.5,0.4
            tx = 0.5
            tl,tm,tu = 0.1,0.5,0.4
        elif transparency == "high":
            t0,t1,t2,t3,t4 = 0,0.7,0.5,0.2,0
            tx = 0.5
            tl,tm,tu = 0.1,0.5,0.1                       
        if hue == 'GRB':
            t0,t1,t2,t3,t4 = t0,t1,t4,t3,t2
        elif hue == "RB" or hue == 'GI':
            tl,tm,tu = tu,tm,tl

        grey = f"rgba(119,136,153,{t0})"
        snow = f"rgba(240,248,255,{t1})"
        cornflower = f"rgba(100,149,237,{t2})"
        crimson = f"rgba(220,20,60,{t3})"
        darkred = f"rgba(100,0,0,{t4})"
        navy = f"rgba(0,0,128,{t4})"

        black = f"rgba(0,0,0,{tx})"
        ghost = f"rgba(248,248,255,{tx})"

        midnight = f"rgba(25,25,112,{tl})"
        maroon = f"rgba(128,0,0,{tu})"

        indigo = f"rgba(75,0,130,{tl})"        
        sea = f"rgba(32,178,170,{tu})"

        slate = f"rgba(40,79,79,{0.7})"        
        fire = f"rgba(178,34,34,{0.7})"  
        alice = f"rgba(240,248,255,{0.7})"  
        rose = f"rgba(255,228,225,{0.7})" 



        
        if hue == "GBR":            
            return [(0, grey), (d0, snow), (d1, cornflower),(d2, crimson),(1.0, darkred)]
        elif hue == "GRB":            
            return [(0, grey), (d0, snow), (d1, crimson),(d2, cornflower),(1.0, navy)]                
        elif hue == "BW":
            return [(0, black),(1.0, ghost)]
        elif hue == "WB":
            return [(0, ghost),(1.0, black)]
        elif hue == "RB":
            if d0 <= 0:
                return [(0,maroon),(0.5,ghost),(1.0,midnight)]
            else:
                return [(0,maroon),(d0,ghost),(1.0,midnight)]
        elif hue == "BR":
            if d0 <= 0:
                return [(0,midnight),(0.5,ghost),(1.0,maroon)]
            else:
                return [(0,midnight),(d0,ghost),(1.0,maroon)]
        elif hue == "IG":
            if d0 <= 0:
                return [(0,indigo),(0.5,ghost),(1.0,sea)]
            else:
                return [(0,indigo),(d0,ghost),(1.0,sea)]
        elif hue == "GI":
            if d0 <= 0:
                return [(0,sea),(0.5,ghost),(1.0,indigo)]
            else:
                return [(0,sea),(d0,ghost),(1.0,indigo)]
        elif hue == "MASK":
            m2 = min(min_r,max_r)
            m3 = max(min_r,max_r)
            m1 = max(0,min(min_r,max_r)-0.01)
            m4 = min(1,max(min_r,max_r)+0.01)
            #return [(0,shell),(m1,shell),(m2,slate),(m3,fire),(m4,shell),(1.0,shell)]
            return [(0,alice),(0.1,alice),(0.2,slate),(0.8,fire),(0.9,rose),(1.0,rose)]                        
            #return [(0,shell),(0.5,slate),(1.0,shell)]
            
            

      

        


        
    

            
    


                    


"""
RSA 4/2/23
https://pynative.com/make-python-class-json-serializable/#:~:text=Use%20toJSON()%20Method%20to%20make%20class%20JSON%20serializable&text=So%20we%20don't%20need,Python%20Object%20to%20JSON%20string.

"""

import os
from os.path import exists
import json
from maptial.xyz import vectorthree as v3

class PdbObject(object):
    def __init__(self, pdb_code):
        # PUBLIC INTERFACE        
        self.pdb_code = pdb_code       
        self.lines = [] # consists of tuple of 
        #ATOM      1  N   GLY A 707   -0003.101   9.984  20.822  1.00  1.81    N  
        #         aid  atm  aa  ch  rid      x         y     z      occ   bf ele     
    
    def get_first_three(self):
        atm1, atm2, atm3 = {},{},{}
        count = 0
        for atm in self.lines:
            if atm["atm"] == "CA" and (atm["version"] == "A" or atm["version"] == ""):
                atm1 = atm
                count += 1
            elif atm["atm"] == "C" and (atm["version"] == "A" or atm["version"] == ""):
                atm2 = atm
                count += 1
            elif atm["atm"] == "O" and (atm["version"] == "A" or atm["version"] == ""):
                atm3 = atm
                count += 1
            if count == 3:
                return atm1,atm2,atm3
        return atm1,atm2,atm3

    def get_coords_key(self,key):
        atm = self.get_atm_key(key)
        if atm == {} or atm == None:
            return "(0,0,0)"
        return f"({atm['x']},{atm['y']},{atm['z']})"

    def get_coords(self,atm):
        if atm == {}:
            return "(0,0,0)"
        return f"({atm['x']},{atm['y']},{atm['z']})"

    def get_atm_key(self,key):
        #A:709@C.A
        try:
            sx = key.split("@")
            s01 = sx[0].split(":")
            s23 = sx[1].split(".")
            ch,rid,am,ver = s01[0],s01[1],s23[0],s23[1]
            for atm in self.lines:        
                if atm["chain"] == ch and atm["rid"] == rid and atm["atm"] == am and atm["version"] == ver:        
                    return atm
        except Exception as e:
            print("Error finding key",e)
            return {}
    
    def get_key(self,atm):
        if atm == {}:
            return ""
        #A:709@C.A
        return f"{atm['chain']}:{atm['rid']}@{atm['atm']}.{atm['version']}"
    
    def get_description(self,atm,dis):
        if atm == {}:
            return ""
        #A:709@C.A
        return f"{round(dis,3)}:{atm['aa']}:{atm['chain']}:{atm['rid']}@{atm['atm']}.{atm['version']}"
    
    def get_next_key(self,key, offset=1):
        try:
            if len(key) < 4:
                return ""
            atm = self.get_atm_key(key)
            ridn = int(atm["rid"]) + offset
            return f"{atm['chain']}:{ridn}@{atm['atm']}.{atm['version']}"
        except:
            return ""
                        
    # if we add lines from a cif file I will do something different    
    def add_line_string(self,line):
        #ATOM    435  OE1AGLU A  23      12.418  13.330   0.541  0.58  7.57           O  
        if "ATOM" in line[:6] or "HETATM" in line[:8]:                        
            aid = line[6:12].strip()
            am = line[12:16].strip()
            ver = line[16:17].strip()
            aa = line[17:21].strip()
            ch = line[21:23].strip()
            rid = line[23:29].strip()
            x = line[29:38].strip()
            y = line[38:46].strip()
            z = line[46:54].strip()
            occ = line[54:60].strip()
            bf = line[60:66].strip()
            ele = line[66:].strip()
            atm = {}
            atm["rid"] = rid
            atm["aid"] = aid
            atm["aa"] = aa
            atm["atm"] = am
            atm["chain"] = ch
            if ver == "":
                ver = "A"
            atm["version"] = ver            
            atm["x"] = x
            atm["y"] = y
            atm["z"] = z
            atm["occupancy"] = occ
            atm["bfactor"] = bf
            atm["element"] = ele        
            self.lines.append(atm)
            #if "HETATM" in line[:8]:
            #    print(line)
            #    print(atm)

    def get_atom_coords(self):
        atoms = []
        for atm in self.lines:
            coord = v3.VectorThree(atm["x"],atm["y"],atm["z"])
            atoms.append(coord)
        return atoms
                
    def get_inscope_atoms(self,anchor,distance,log_level=0):
        in_scope = []
        for atm in self.lines:
            crd = v3.VectorThree(float(atm["x"]),float(atm["y"]),float(atm["z"]))
            dis = anchor.distance(crd)
            if dis <= distance:
                in_scope.append(atm)
                if log_level > 0:                        
                    print("neighbour in scope", self.get_description(atm,dis))
        return in_scope

    def get_neighbours(self,coord,rnge,atoms=None,ishtml=False):
        a = ""
        if atoms == None:
            atoms = self.lines
        for atm in atoms:
            crd = v3.VectorThree(float(atm["x"]),float(atm["y"]),float(atm["z"]))
            dis = coord.distance(crd)
            if dis >= rnge[0] and dis <= rnge[1]:
                desc = self.get_description(atm,dis)
                if desc not in a:
                    if ishtml:
                        a+="<br>......"+ desc
                    else:
                        a+="\n......"+desc
        return a

    def toJson(self):
        return json.dumps(self, default=lambda o: o.__dict__)
    
    def fromJson(self, jsndic):    
        self.pdb_code = jsndic["pdb_code"]
        self.lines = jsndic["lines"]
        
        
                                
    


"""
RSA 25/2/23

This class handles 3d matrices
"""

import math
import numpy as np

# class interface
from . import vectorthree as v3

class MatrixThree(object):
    def __init__(self, vals = [0.,0.,0.,0.,0.,0.,0.,0.,0.]):                
        self.npy = np.array(vals)        
        self.npy = self.npy.reshape(3, 3)  
                                
    def get_inverse(self):
        ############################### https://numpy.org/doc/stable/reference/generated/numpy.linalg.inv.html        
        try:
            from numpy.linalg import inv   
            from numpy.linalg import det
            minv = inv(self.npy)
            m3inv = MatrixThree(minv.reshape(9))
            return m3inv
        except:
            print("!!!!!! error", self.get_key(), det(self.npy))
        return None
        ###############################
                
    def get_determinant(self):                
        #################################
        from numpy.linalg import det
        return det(self.npy)
        #################################
                        
    def get_value(self, row, col):                
        return self.npy[row,col]
                                    
    def put_value(self, val, row, col):            
        self.npy[row,col] = val
                
    def multiply(self, col, byRow):
        ###################################### https://numpy.org/doc/stable/reference/generated/numpy.matmul.html        
        if byRow:
            mm= np.matmul(self.npy.transpose(), col.npy)
        else:
            mm= np.matmul(self.npy, col.npy)
                        
        vv = mm.reshape(3)
        #print("ToList", vv.tolist())
        vm3 = v3.VectorThree(vv[0],vv[1],vv[2])
        return vm3
        
    def get_key(self,rnd=-1):        
        if rnd > -1:                        
            return str(self.npy.reshape(9).round(rnd))
        else:
            return str(self.npy.reshape(9))

    
        